#===============================================================================
# Copyright 2010 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""
This module contains classes implementing exclusive locks for platforms with
fcntl (UNIX and Mac OS X) and Windows. Whoosh originally used directory creation
as a locking method, but it had the problem that if the program crashed the
lock directory was left behind and would keep the index locked until it was
cleaned up. Using OS-level file locks fixes this.
"""

import errno
import os
import time


def try_for(fn, timeout=5.0, delay=0.1):
    """Calls ``fn`` every ``delay`` seconds until it returns True or ``timeout``
    seconds elapse. Returns True if the lock was acquired, or False if the
    timeout was reached.

    :param timeout: Length of time (in seconds) to keep retrying to acquire the
        lock. 0 means return immediately. Only used when blocking is False.
    :param delay: How often (in seconds) to retry acquiring the lock during
        the timeout period. Only used when blocking is False and timeout > 0.
    """
    
    until = time.time() + timeout
    v = fn()
    while not v and time.time() < until:
        time.sleep(delay)
        v = fn()
    return v


class LockBase(object):
    """Base class for file locks.
    """
    
    def __init__(self, filename):
        self.fd = None
        self.filename = filename
        self.locked = False
    
    def __del__(self):
        if hasattr(self, "fd") and self.fd:
            try:
                self.release()
            except:
                pass
    
    def acquire(self, blocking=False):
        """Acquire the lock. Returns True if the lock was acquired.
        
        :param blocking: if True, call blocks until the lock is acquired.
            This may not be available on all platforms. On Windows, this is
            actually just a delay of 10 seconds, rechecking every second.
        """
        pass
    
    def release(self):
        pass
    
    
class FcntlLock(LockBase):
    """File lock based on UNIX-only fcntl module.
    """
    
    def acquire(self, blocking=False):
        import fcntl
        
        flags = os.O_CREAT | os.O_WRONLY
        self.fd = os.open(self.filename, flags)
        
        mode = fcntl.LOCK_EX
        if not blocking:
            mode |= fcntl.LOCK_NB
        
        try:
            fcntl.flock(self.fd, mode)
            self.locked = True
            return True
        except IOError, e:
            if e.errno not in (errno.EAGAIN, errno.EACCES):
                raise
            os.close(self.fd)
            self.fd = None
            return False
        
    def release(self):
        import fcntl
        fcntl.flock(self.fd, fcntl.LOCK_UN)
        os.close(self.fd)
        self.fd = None


class MsvcrtLock(LockBase):
    """File lock based on Windows-only msvcrt module.
    """
    
    def acquire(self, blocking=False):
        import msvcrt
        
        flags = os.O_CREAT | os.O_WRONLY
        mode = msvcrt.LK_NBLCK
        if blocking:
            mode = msvcrt.LK_LOCK
            
        self.fd = os.open(self.filename, flags)
        try:
            msvcrt.locking(self.fd, mode, 1)
            return True
        except IOError, e:
            if e.errno not in (errno.EAGAIN, errno.EACCES, errno.EDEADLK):
                raise
            os.close(self.fd)
            self.fd = None
            return False
        
    def release(self):
        if self.fd is None:
            raise Exception("Lock was not acquired")
        
        import msvcrt
        msvcrt.locking(self.fd, msvcrt.LK_UNLCK, 1)
        os.close(self.fd)
        self.fd = None


if os.name == "nt":
    FileLock = MsvcrtLock
else:
    FileLock = FcntlLock






from __future__ import with_statement
import unittest
import os.path
from random import random, randint

from whoosh.formats import (Characters, CharacterBoosts, DocBoosts, Existence,
                            Frequency, Positions, PositionBoosts)
from whoosh.filedb.filepostings import FilePostingWriter, FilePostingReader
from whoosh.util import float_to_byte, byte_to_float
from whoosh.support.testing import TempStorage


class TestPostings(unittest.TestCase):
    def make_postings(self):
        postings = [(1, 23), (3, 45), (12, 2), (34, 21), (43, 7), (67, 103), (68, 1), (102, 31),
                    (145, 4), (212, 9), (283, 30), (291, 6), (412, 39), (900, 50), (905, 28), (1024, 8),
                    (1800, 13), (2048, 3), (15000, 40)]
        return postings
    
    def test_readwrite(self):
        with TempStorage("readwrite") as st:
            format = Frequency(None)
            postings = self.make_postings()
            
            postfile = st.create_file("readwrite")
            fpw = FilePostingWriter(postfile, blocklimit=8)
            fpw.start(format)
            for id, freq in postings:
                fpw.write(id, float(freq), format.encode(freq), 0)
            fpw.close()
            
            postfile = st.open_file("readwrite")
            fpr = FilePostingReader(postfile, 0, format)
            self.assertEqual(postings, list(fpr.items_as("frequency")))
            postfile.close()
        
    def test_skip(self):
        with TempStorage("skip") as st:
            format = Frequency(None)
            postings = self.make_postings()
            
            postfile = st.create_file("skip")
            fpw = FilePostingWriter(postfile, blocklimit=8)
            fpw.start(format)
            for id, freq in postings:
                fpw.write(id, float(freq), format.encode(freq), 0)
            fpw.close()
            
            postfile = st.open_file("skip")
            fpr = FilePostingReader(postfile, 0, format)
            fpr.skip_to(220)
            self.assertEqual(postings[10:], list(fpr.items_as("frequency")))
            postfile.close()
    
    def roundtrip(self, postings, format, astype):
        with TempStorage("roundtrip") as st:
            postfile = st.create_file(astype)
            getweight = format.decoder("weight")
            fpw = FilePostingWriter(postfile, blocklimit=8)
            fpw.start(format)
            for id, value in postings:
                v = format.encode(value)
                fpw.write(id, getweight(v), v, 0)
            fpw.close()
            
            postfile = st.open_file(astype)
            fpr = FilePostingReader(postfile, 0, format)
            readback = list(fpr.items_as(astype))
            postfile.close()
            return readback
    
    def test_existence_postings(self):
        postings = []
        docnum = 0
        for _ in xrange(0, 20):
            docnum += randint(1, 10)
            postings.append((docnum, 1))
        
        self.assertEqual(postings, self.roundtrip(postings, Existence(None), "frequency"))
    
    def test_docboost_postings(self):
        postings = []
        docnum = 0
        for _ in xrange(0, 20):
            docnum += randint(1, 10)
            freq = randint(1, 1000)
            boost = byte_to_float(float_to_byte(random() * 2))
            postings.append((docnum, (freq, boost)))
        
        self.assertEqual(postings, self.roundtrip(postings, DocBoosts(None), "docboosts"))
        
    def test_position_postings(self):
        postings = []
        docnum = 0
        for _ in xrange(0, 20):
            docnum += randint(1, 10)
            posns = []
            pos = 0
            for __ in xrange(0, randint(1, 10)):
                pos += randint(1, 10)
                posns.append(pos)
            postings.append((docnum, posns))
        
        self.assertEqual(postings, self.roundtrip(postings, Positions(None), "positions"))
        
        as_freq = [(docnum, len(posns)) for docnum, posns in postings]
        self.assertEqual(as_freq, self.roundtrip(postings, Positions(None), "frequency"))
        
    def test_character_postings(self):
        postings = []
        docnum = 0
        for _ in xrange(0, 20):
            docnum += randint(1, 10)
            posns = []
            pos = 0
            endchar = 0
            for __ in xrange(0, randint(1, 10)):
                pos += randint(1, 10)
                startchar = endchar + randint(3, 10)
                endchar = startchar + randint(3, 10)
                posns.append((pos, startchar, endchar))
            postings.append((docnum, posns))
            
        self.assertEqual(postings, self.roundtrip(postings, Characters(None), "characters"))
        
        as_posns = [(docnum, [pos for pos, sc, ec in posns]) for docnum, posns in postings]
        self.assertEqual(as_posns, self.roundtrip(postings, Characters(None), "positions"))
        
        as_freq = [(docnum, len(posns)) for docnum, posns in as_posns]
        self.assertEqual(as_freq, self.roundtrip(postings, Characters(None), "frequency"))
        
    def test_posboost_postings(self):
        postings = []
        docnum = 0
        for _ in xrange(0, 3):
            docnum += randint(1, 10)
            posns = []
            pos = 0
            for __ in xrange(0, randint(1, 3)):
                pos += randint(1, 10)
                boost = byte_to_float(float_to_byte(random() * 2))
                posns.append((pos, boost))
            postings.append((docnum, posns))
        
        self.assertEqual(postings, self.roundtrip(postings, PositionBoosts(None), "position_boosts"))
        
        as_posns = [(docnum, [pos for pos, boost in posns]) for docnum, posns in postings]
        self.assertEqual(as_posns, self.roundtrip(postings, PositionBoosts(None), "positions"))
        
        as_freq = [(docnum, len(posns)) for docnum, posns in postings]
        self.assertEqual(as_freq, self.roundtrip(postings, PositionBoosts(None), "frequency"))

    def test_charboost_postings(self):
        postings = []
        docnum = 0
        for _ in xrange(0, 20):
            docnum += randint(1, 10)
            posns = []
            pos = 0
            endchar = 0
            for __ in xrange(0, randint(1, 10)):
                pos += randint(1, 10)
                startchar = endchar + randint(3, 10)
                endchar = startchar + randint(3, 10)
                boost = byte_to_float(float_to_byte(random() * 2))
                posns.append((pos, startchar, endchar, boost))
            postings.append((docnum, posns))
        
        self.assertEqual(postings, self.roundtrip(postings, CharacterBoosts(None), "character_boosts"))
        
        as_chars = [(docnum, [(pos, sc, ec) for pos, sc, ec, bst in posns]) for docnum, posns in postings]
        self.assertEqual(as_chars, self.roundtrip(postings, CharacterBoosts(None), "characters"))
        
        as_posbsts = [(docnum, [(pos, bst) for pos, sc, ec, bst in posns]) for docnum, posns in postings]
        self.assertEqual(as_posbsts, self.roundtrip(postings, CharacterBoosts(None), "position_boosts"))
        
        as_posns = [(docnum, [pos for pos, sc, ec, bst in posns]) for docnum, posns in postings]
        self.assertEqual(as_posns, self.roundtrip(postings, CharacterBoosts(None), "positions"))
        
        as_freq = [(docnum, len(posns)) for docnum, posns in as_posns]
        self.assertEqual(as_freq, self.roundtrip(postings, CharacterBoosts(None), "frequency"))

    
        
        


if __name__ == '__main__':
    unittest.main()





#===============================================================================
# Copyright 2010 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""
This module contains support classes for the query parser. These objects are
used to construct the parsed syntax tree of the query. The syntax tree is then
tranlsated into a query tree by calling ``SyntaxObject.query()`` on the object
at the top of the tree.
"""

from whoosh import query
from whoosh.qparser.common import rcompile


class SyntaxObject(object):
    """An object representing parsed text. These objects generally correspond
    to a query object type, and are intermediate objects used to represent the
    syntax tree parsed from a query string, and then generate a query tree from
    the syntax tree. There will be syntax objects that do not have a
    corresponding query type, such as the syntax object representing
    whitespace.
    """
    
    def set_fieldname(self, name, force=False):
        """Returns a version of this syntax object with the field name set to
        the given name. Normally this only changes the field name if the
        field name is not already set, but if the ``force`` keyword argument
        is True, the field name will be changed regardless.
        
        This method is mis-named and confusing, but is used by the parser
        to assign field names to branches of the syntax tree, but only for
        syntax objects that didn't have an explicit field name set by the user.
        """
        
        raise NotImplementedError
    
    def query(self, parser):
        """Returns a query object tree representing this parser object.
        """
        
        raise NotImplementedError


# Grouping objects

class Group(SyntaxObject):
    """Represents a group of syntax objects. These generally correspond to
    compound query objects such as ``query.And`` and ``query.Or``.
    """
    
    # Whether this group can have any number of children. Other than AND and
    # OR, most groups will represent binary queries, so the default is False.
    many = False
    
    # Sub-classes that want to use the default query() implementation should
    # set this to the query class corresponding to this group
    qclass = None
    
    def __init__(self, tokens=None, boost=1.0):
        self.tokens = tokens or []
        self.boost = boost
    
    def __repr__(self):
        r = "%s(%r)" % (self.__class__.__name__, self.tokens)
        if self.boost != 1.0:
            r += "^%s" % self.boost
        return r
    
    def __nonzero__(self):
        return bool(self.tokens)
    
    def __iter__(self):
        return iter(self.tokens)
    
    def __len__(self):
        return len(self.tokens)
    
    def __getitem__(self, n):
        return self.tokens.__getitem__(n)
    
    def __setitem__(self, n, v):
        self.tokens.__setitem__(n, v)
    
    def __delitem__(self, n):
        self.tokens.__delitem__(n)
    
    def insert(self, n, v):
        self.tokens.insert(n, v)
    
    def set_boost(self, b):
        return self.__class__(self.tokens[:], boost=b)
    
    def set_fieldname(self, name, force=False):
        return self.__class__([t.set_fieldname(name, force)
                               for t in self.tokens])
    
    def append(self, item):
        self.tokens.append(item)
        
    def extend(self, items):
        self.tokens.extend(items)
    
    def pop(self):
        return self.tokens.pop()
    
    def reverse(self):
        self.tokens.reverse()
    
    def query(self, parser):
        return self.qclass([t.query(parser) for t in self.tokens],
                           boost=self.boost)
        
    def empty(self):
        return self.__class__(boost=self.boost)


class AndGroup(Group):
    """Syntax group corresponding to an And query.
    """
    
    # This group can have more than 2 children
    many = True
    qclass = query.And
    

class OrGroup(Group):
    """Syntax group corresponding to an Or query.
    """
    
    # This group can have more than 2 children
    many = True
    qclass = query.Or


class AndNotGroup(Group):
    """Syntax group corresponding to an AndNot query.
    """
    
    def query(self, parser):
        assert len(self.tokens) == 2
        return query.AndNot(self.tokens[0].query(parser),
                            self.tokens[1].query(parser), boost=self.boost)
    

class AndMaybeGroup(Group):
    """Syntax group corresponding to an AndMaybe query.
    """
    
    def query(self, parser):
        assert len(self.tokens) == 2
        return query.AndMaybe(self.tokens[0].query(parser),
                              self.tokens[1].query(parser), boost=self.boost)


class RequireGroup(Group):
    """Syntax group corresponding to a Require query.
    """
    
    def query(self, parser):
        assert len(self.tokens) == 2, self.tokens
        return query.Require(self.tokens[0].query(parser),
                             self.tokens[1].query(parser), boost=self.boost)


class OrderedGroup(Group):
    """Syntax group corresponding to the Ordered query.
    """
    
    many = True
    qclass = query.Ordered


class DisMaxGroup(Group):
    """Syntax group corresponding to a DisjunctionMax query.
    """
    
    def __init__(self, tokens=None, tiebreak=0.0, boost=None):
        super(DisMaxGroup, self).__init__(tokens)
        self.tiebreak = tiebreak
    
    def __repr__(self):
        r = "dismax(%r" % self.tokens
        if self.tiebreak != 0:
            r += " tb=%s" % self.tiebreak
        r += ")"
        return r
    
    def query(self, parser):
        return query.DisjunctionMax([t.query(parser) for t in self.tokens],
                                    tiebreak=self.tiebreak)
        
    def empty(self):
        return self.__class__(tiebreak=self.tiebreak, boost=self.boost)


class NotGroup(Group):
    """Syntax group corresponding to a Not query.
    """
    
    def __repr__(self):
        return "NOT(%r)" % self.tokens
    
    def query(self, parser):
        assert len(self.tokens) == 1
        return query.Not(self.tokens[0].query(parser))
    

# Parse-able tokens

class Token(SyntaxObject):
    """A parse-able token object. Each token class has an ``expr`` attribute
    containing a regular expression that matches the token text. When this
    expression is found, the class/object's ``create()`` method is called and
    returns a token object to represent the match in the token stream.
    
    Many token classes will do the parsing using class methods and put
    instances of themselves in the token stream, however parseable objects
    requiring configuration (such as the :class:`Operator` subclasses may use
    separate objects for doing the parsing and embodying the token.
    """
    
    fieldname = None
    endpos = None
    
    def set_boost(self, b):
        return self
    
    def set_fieldname(self, name, force=False):
        return self
    
    @classmethod
    def match(cls, text, pos):
        return cls.expr.match(text, pos)
    
    @classmethod
    def create(cls, parser, match):
        return cls()
    
    def query(self, parser):
        raise NotImplementedError


class Operator(Token):
    """Represents a search operator which modifies the token stream by putting
    certain tokens into a :class:`Group` object. For example, an "and" infix
    operator would put the two tokens on either side of the operator into
    an :class:`AndGroup`.
    
    This is the base class for operators. Subclasses must implement the
    :meth:`Operator.make_group` method.
    """
    
    def __init__(self, expr, grouptype, left_assoc=True):
        """
        :param expr: a pattern string or compiled expression of the token text.
        :param grouptype: a :class:`Group` subclass that should be created to
            contain objects affected by the operator.
        """
        
        self.expr = rcompile(expr)
        self.grouptype = grouptype
        self.left_assoc = left_assoc
    
    def __repr__(self):
        return "%s<%s>" % (self.__class__.__name__, self.expr.pattern)
    
    def make_group(self, parser, stream, position):
        raise NotImplementedError
    
    def match(self, text, pos):
        return self.expr.match(text, pos)
    
    def create(self, parser, match):
        return self
    
    
class PrefixOperator(Operator):
    """Implements a prefix operator. That is, the token immediately following
    the operator will be put into the group.
    """
    
    def make_group(self, parser, stream, position):
        if position < len(stream) - 1:
            del stream[position]
            stream[position] = self.grouptype([stream[position]])
        else:
            del stream[position]
        return position
    
    
class PostfixOperator(Operator):
    """Implements a postfix operator. That is, the token immediately preceding
    the operator will be put into the group.
    """
    
    def make_group(self, parser, stream, position):
        if position > 0:
            del stream[position]
            stream[position - 1] = self.grouptype([stream[position - 1]])
        else:
            del stream[position]
        return position


class InfixOperator(Operator):
    """Implements an infix operator. That is, the tokens immediately on either
    side of the operator will be put into the group.
    """
    
    def __init__(self, expr, grouptype, left_assoc=True):
        """
        :param expr: a pattern string or compiled expression of the token text.
        :param grouptype: a :class:`Group` subclass that should be created to
            contain objects affected by the operator.
        :param left_assoc: if True, the operator is left associative. Otherwise
            it is right associative.
        """
        
        super(InfixOperator, self).__init__(expr, grouptype)
        self.left_assoc = left_assoc
    
    def make_group(self, parser, stream, position):
        if position > 0 and position < len(stream) - 1:
            left = stream[position - 1]
            right = stream[position + 1]
            
            # The first two clauses check whether the "strong" side is already
            # a group of the type we are going to create. If it is, we just
            # append the "weak" side to the "strong" side instead of creating
            # a new group inside the existing one. This is necessary because
            # we can quickly run into Python's recursion limit otherwise.
            if self.grouptype.many and self.left_assoc and isinstance(left, self.grouptype):
                left.append(right)
                del stream[position:position + 2]
            elif self.grouptype.many and not self.left_assoc and isinstance(right, self.grouptype):
                right.insert(0, left)
                del stream[position - 1:position + 1]
                return position - 1
            else:
                # Replace the operator and the two surrounding objects
                stream[position - 1:position + 2] = [self.grouptype([left, right])]
        else:
            del stream[position]
        return position


class Singleton(Token):
    """Base class for tokens that don't carry any information specific to
    each instance (e.g. "open parenthesis" token), so they can all share the
    same instance.
    """
    
    me = None
    
    def __repr__(self):
        return self.__class__.__name__
    
    @classmethod
    def create(cls, parser, match):
        if not cls.me:
            cls.me = cls()
        return cls.me


class White(Singleton):
    expr = rcompile("\\s+")
    

class ErrorToken(Token):
    """A token representing an unavoidable parsing error. The ``query()``
    method always returns NullQuery.
    
    The default parser usually does not produce "errors" (text that doesn't
    match the syntax is simply treated as part of the query), so this is mostly
    for use by plugins that may add more restrictive parsing, for example
    :class:`DateParserPlugin`.
    
    Since the corresponding NullQuery will be filtered out when the query is
    normalized, this is really only useful for debugging and possibly for
    plugin filters.
    
    The ``token`` attribute may contain the token that produced the error.
    """
    
    def __init__(self, token):
        self.token = token
        
    def __repr__(self):
        return "<%s (%r)>" % (self.__class__.__name__, self.token)
    
    def query(self, parser):
        return query.NullQuery


class BasicSyntax(Token):
    """Base class for "basic" (atomic) syntax -- term, prefix, wildcard,
    phrase, range.
    """
    
    expr = None
    qclass = None
    tokenize = False
    removestops = False
    
    def __init__(self, text, fieldname=None, boost=1.0):
        self.fieldname = fieldname
        self.text = text
        self.boost = boost
    
    def set_boost(self, b):
        return self.__class__(self.text, fieldname=self.fieldname, boost=b)
    
    def set_fieldname(self, name, force=False):
        if force or self.fieldname is None:
            return self.__class__(self.text, fieldname=name, boost=self.boost)
        else:
            return self
    
    def __repr__(self):
        r = "%s:%r" % (self.fieldname, self.text)
        if self.boost != 1.0:
            r += "^%s" % self.boost
        return r
    
    @classmethod
    def create(cls, parser, match):
        return cls(match.group(0))
    
    def query(self, parser):
        fieldname = self.fieldname or parser.fieldname
        termclass = self.qclass or parser.termclass
        
        return parser.term_query(fieldname, self.text, termclass,
                                 boost=self.boost, tokenize=self.tokenize,
                                 removestops=self.removestops)
        

class Word(BasicSyntax):
    """Syntax object representing a term.
    """
    
    expr = rcompile("[^ \t\r\n)]+")
    tokenize = True
    removestops = True
 


from __future__ import with_statement
import unittest

from whoosh import analysis, fields, formats, query
from whoosh.ramindex import RamIndex
from whoosh.support.testing import TempIndex


class TestRamIndex(unittest.TestCase):
    def make_index(self):
        ana = analysis.StandardAnalyzer(stoplist=None)
        sc = fields.Schema(id=fields.ID(stored=True),
                           text=fields.TEXT(analyzer=ana, vector=formats.Frequency(ana)),
                           subs=fields.NUMERIC(int, stored=True))
        ix = RamIndex(sc)
        ix.add_document(id=u"fieldtype",
                        text=u"The FieldType object supports the following attributes",
                        subs=56)
        ix.add_document(id=u"format",
                        text=u"the storage format for the field contents",
                        subs=100)
        ix.add_document(id=u"vector",
                        text=u"the storage format for the field vectors (forward index)",
                        subs=23)
        ix.add_document(id=u"scorable",
                        text=u"whether searches against this field may be scored.",
                        subs=34)
        ix.add_document(id=u"stored",
                        text=u"whether the content of this field is stored for each document.",
                        subs=575)
        ix.add_document(id=u"unique",
                        text=u"whether this field value is unique to each document.",
                        subs=2)
        ix.add_document(id=u"const",
                        text=u"The constructor for the base field type simply",
                        subs=58204)
        return ix
    
    def test_docs_method(self):
        ix = self.make_index()
        with ix.searcher() as s:
            self.assertEqual(s.document(id="vector"), {"id": "vector", "subs": 23})
            self.assertEqual(s.document(subs=2), {"id": "unique", "subs": 2})
    
    def test_searching(self):
        with self.make_index().searcher() as s:
            def _runq(q, result, **kwargs):
                r = s.search(q, scored=False, **kwargs)
                self.assertEqual([d["id"] for d in r], result)
            
            _runq(query.Term("text", u"format"), ["format", "vector"])
            _runq(query.Term("text", u"the"), ["fieldtype", "format", "vector", "stored", "const"])
            _runq(query.Prefix("text", u"st"), ["format", "vector", "stored"])
            _runq(query.Wildcard("id", u"*st*"), ["stored", "const"])
            _runq(query.TermRange("id", u"c", u"s"), ["fieldtype", "format", "const"])
            _runq(query.NumericRange("subs", 10, 100), ["fieldtype", "format", "vector", "scorable"])
            _runq(query.Phrase("text", ["this", "field"]), ["scorable", "stored", "unique"], limit=None)
            _runq(query.Every(), ["fieldtype", "format", "vector", "scorable", "stored", "unique", "const"])
            _runq(query.Every("subs"), ["fieldtype", "format", "vector", "scorable", "stored", "unique", "const"])
    
    def test_update(self):
        schema = fields.Schema(id=fields.ID(unique=True, stored=True),
                               text=fields.ID(stored=True))
        ix = RamIndex(schema)
        for word in u"alfa bravo charlie delta".split():
            ix.update_document(id=word[0], text=word)
        for word in u"apple burrito cat dollhouse".split():
            ix.update_document(id=word[0], text=word)
        
        self.assertTrue(ix.has_deletions())
        self.assertEqual(ix.deleted, set([0, 1, 2, 3]))
        self.assertEqual(ix.doc_count(), 4)
        self.assertEqual([(d["id"], d["text"]) for d in ix.all_stored_fields()],
                         [("a", "apple"), ("b", "burrito"), ("c", "cat"), ("d", "dollhouse")])
    
    def test_contains(self):
        ix = self.make_index()
        totext = ix.schema["subs"].to_text
        self.assertTrue(("text", u"format") in ix)
        self.assertTrue(("id", u"fieldtype") in ix)
        self.assertTrue(("subs", totext(23)) in ix)
        self.assertFalse(("foo", u"bar") in ix)
        self.assertFalse(("text", u"repeat") in ix)
        self.assertFalse(("id", u"galonk") in ix)
        self.assertFalse(("subs", totext(500)) in ix)
    
    def test_indexing(self):
        ix = self.make_index()
        with ix.searcher() as s:
            q = query.Term("text", "format")
            r = s.search(q)
            self.assertEqual(len(r), 2)
            self.assertEqual(r[0]["id"], "format")
            self.assertEqual(r[0]["subs"], 100)
            self.assertEqual(r[1]["id"], "vector")
            self.assertEqual(r[1]["subs"], 23)

    def test_delete_doc(self):
        ix = self.make_index()
        ix.delete_document(2)
        ix.delete_document(5)
        with ix.searcher() as s:
            self.assertEqual([d["id"] for d in s.search(query.Every())],
                             ["fieldtype", "format", "scorable", "stored", "const"])

    def test_stored(self):
        r = self.make_index().reader()
        
        self.assertEqual(r.stored_fields(2), {"id": "vector", "subs": 23})
        
        target = [{"id": "fieldtype", "subs": 56},
                  {"id": "format", "subs": 100},
                  {"id": "vector", "subs": 23},
                  {"id": "scorable", "subs": 34},
                  {"id": "stored", "subs": 575},
                  {"id": "unique", "subs": 2},
                  {"id": "const", "subs": 58204},
                  ]
        self.assertEqual(list(r.all_stored_fields()), target)

    def test_field_length(self):
        r = self.make_index().reader()
        self.assertEqual(r.field_length("id"), 0)  # Not scorable
        self.assertEqual(r.max_field_length("id"), 0)
        self.assertEqual(r.field_length("subs"), 0)
        self.assertEqual(r.max_field_length("subs"), 0)
        self.assertEqual(r.field_length("text"), 59)
        self.assertEqual(r.max_field_length("text"), 11)
        self.assertEqual(r.doc_field_length(3, "text"), 8)
        
        self.assertEqual(r.field_length("text"), 59)
        self.assertEqual(r.max_field_length("text"), 11)
        self.assertEqual(r.doc_field_length(3, "text"), 8)
        
        self.assertEqual(r.doc_frequency("text", "the"), 5)
        self.assertEqual(r.frequency("text", "the"), 9)

    def test_deleting(self):
        ix = self.make_index()
        ix.delete_by_term("id", u"vector")
        
        self.assertTrue(ix.has_deletions())
        
        with ix.searcher() as s:
            q = query.Term("text", "format")
            r = s.search(q)
            self.assertEqual(len(r), 1)
            self.assertEqual(r[0]["id"], "format")
            self.assertEqual(r[0]["subs"], 100)
        
    def test_iter(self):
        r = self.make_index().reader()
        
        everything = [("id", u'const', 1, 1), ("id", u'fieldtype', 1, 1), ("id", u'format', 1, 1),
                      ("id", u'scorable', 1, 1), ("id", u'stored', 1, 1), ("id", u'unique', 1, 1),
                      ("id", u'vector', 1, 1),  ("text", u'against', 1, 1),
                      ("text", u'attributes', 1, 1), ("text", u'base', 1, 1), ("text", u'be', 1, 1),
                      ("text", u'constructor', 1, 1), ("text", u'content', 1, 1), ("text", u'contents', 1, 1),
                      ("text", u'document', 2, 2), ("text", u'each', 2, 2), ("text", u'field', 6, 6),
                      ("text", u'fieldtype', 1, 1), ("text", u'following', 1, 1), ("text", u'for', 4, 4),
                      ("text", u'format', 2, 2), ("text", u'forward', 1, 1), ("text", u'index', 1, 1),
                      ("text", u'is', 2, 2), ("text", u'may', 1, 1), ("text", u'object', 1, 1),
                      ("text", u'of', 1, 1), ("text", u'scored', 1, 1), ("text", u'searches', 1, 1),
                      ("text", u'simply', 1, 1), ("text", u'storage', 2, 2), ("text", u'stored', 1, 1),
                      ("text", u'supports', 1, 1), ("text", u'the', 5, 9), ("text", u'this', 3, 3),
                      ("text", u'to', 1, 1), ("text", u'type', 1, 1), ("text", u'unique', 1, 1),
                      ("text", u'value', 1, 1), ("text", u'vectors', 1, 1), ("text", u'whether', 3, 3)]
        
        self.assertEqual([item for item in r if item[0] != 'subs'], everything)
        self.assertEqual(list(r.iter_from("text", u"su")), everything[32:])
        self.assertEqual(list(r.lexicon("text")), [x[1] for x in everything if x[0] == "text"])
        self.assertEqual(list(r.iter_field("text")), [x[1:] for x in everything if x[0] == "text"])
        self.assertEqual(list(r.iter_field("text", "st")), [x[1:] for x in everything if x[0] == "text" and x[1] >= "st"])
    
    def test_vectors(self):
        ix = self.make_index()
        r = ix.reader()
        
        self.assertFalse(r.has_vector(0, "id"))
        self.assertTrue(r.has_vector(0, "text"))
        
        target = [(u'contents', 1), (u'field', 1), (u'for', 1), (u'format', 1),
                  (u'storage', 1), (u'the', 2)]
        vec = list(r.vector_as("frequency", 1, "text"))
        self.assertEqual(target, vec)
    
    def test_todisk(self):
        ix = self.make_index()

        with TempIndex(ix.schema, "ramtodisk") as fix:
            w = fix.writer()
            w.add_reader(ix.reader())
            w.commit()
            
    def test_threaded(self):
        from threading import Thread
        
        class TWriter(Thread):
            def __init__(self, ix):
                Thread.__init__(self)
                self.ix = ix
                
            def run(self):
                ix = self.ix
                for i in xrange(1000):
                    ix.update_document(id=unicode(i), key=u"a")
        
        class TReader(Thread):
            def __init__(self, ix):
                Thread.__init__(self)
                self.ix = ix
                self.go = True
            
            def run(self):
                s = self.ix.searcher()
                while self.go:
                    r = s.search(query.Term("key", u"a"))
                    assert len(r) == 1
        
        schema = fields.Schema(id=fields.ID(stored=True),key=fields.ID(unique=True, stored=True))
        ix = RamIndex(schema)
        tw = TWriter(ix)
        tr = TReader(ix)
        tw.start()
        tr.start()
        tw.join()
        tr.go = False
        tr.join()
        
        self.assertEqual(ix.doc_count(), 1)
        with ix.searcher() as s:
            self.assertEqual(len(list(s.documents(key="a"))), 1)

        
if __name__ == '__main__':
    unittest.main()




from __future__ import with_statement
import unittest

from whoosh import analysis, fields, formats, qparser, query, searching
from whoosh.filedb.filestore import RamStorage
from whoosh.util import permutations


class TestResults(unittest.TestCase):
    def test_score_retrieval(self):
        schema = fields.Schema(title=fields.TEXT(stored=True),
                               content=fields.TEXT(stored=True))
        storage = RamStorage()
        ix = storage.create_index(schema)
        writer = ix.writer()
        writer.add_document(title=u"Miss Mary",
                            content=u"Mary had a little white lamb its fleece was white as snow")
        writer.add_document(title=u"Snow White",
                            content=u"Snow white lived in the forest with seven dwarfs")
        writer.commit()
        
        with ix.searcher() as s:
            results = s.search(query.Term("content", "white"))
            self.assertEqual(len(results), 2)
            self.assertEqual(results[0]['title'], u"Miss Mary")
            self.assertEqual(results[1]['title'], u"Snow White")
            self.assertNotEqual(results.score(0), None)
            self.assertNotEqual(results.score(0), 0)
            self.assertNotEqual(results.score(0), 1)

    def test_resultcopy(self):
        schema = fields.Schema(a=fields.TEXT(stored=True))
        st = RamStorage()
        ix = st.create_index(schema)
        
        w = ix.writer()
        w.add_document(a=u"alfa bravo charlie")
        w.add_document(a=u"bravo charlie delta")
        w.add_document(a=u"charlie delta echo")
        w.add_document(a=u"delta echo foxtrot")
        w.commit()
        
        with ix.searcher() as s:
            r = s.search(qparser.QueryParser("a", None).parse(u"charlie"))
            self.assertEqual(len(r), 3)
            rcopy = r.copy()
            self.assertEqual(r.top_n, rcopy.top_n)
        
    def test_resultslength(self):
        schema = fields.Schema(id=fields.ID(stored=True),
                               value=fields.TEXT)
        st = RamStorage()
        ix = st.create_index(schema)
        
        w = ix.writer()
        w.add_document(id=u"1", value=u"alfa alfa alfa alfa alfa")
        w.add_document(id=u"2", value=u"alfa alfa alfa alfa")
        w.add_document(id=u"3", value=u"alfa alfa alfa")
        w.add_document(id=u"4", value=u"alfa alfa")
        w.add_document(id=u"5", value=u"alfa")
        w.add_document(id=u"6", value=u"bravo")
        w.commit()
        
        with ix.searcher() as s:
            r = s.search(query.Term("value", u"alfa"), limit=3)
            self.assertEqual(len(r), 5)
            self.assertEqual(r.scored_length(), 3)
            self.assertEqual(r[10:], [])

    def test_pages(self):
        from whoosh.scoring import Frequency
        
        schema = fields.Schema(id=fields.ID(stored=True), c=fields.TEXT)
        ix = RamStorage().create_index(schema)
        
        w = ix.writer()
        w.add_document(id=u"1", c=u"alfa alfa alfa alfa alfa alfa")
        w.add_document(id=u"2", c=u"alfa alfa alfa alfa alfa")
        w.add_document(id=u"3", c=u"alfa alfa alfa alfa")
        w.add_document(id=u"4", c=u"alfa alfa alfa")
        w.add_document(id=u"5", c=u"alfa alfa")
        w.add_document(id=u"6", c=u"alfa")
        w.commit()
        
        with ix.searcher(weighting=Frequency) as s:
            q = query.Term("c", u"alfa")
            r = s.search(q)
            self.assertEqual([d["id"] for d in r], ["1", "2", "3", "4", "5", "6"])
            r = s.search_page(q, 2, pagelen=2)
            self.assertEqual([d["id"] for d in r], ["3", "4"])
            
            r = s.search_page(q, 2, pagelen=4)
            self.assertEqual(r.total, 6)
            self.assertEqual(r.pagenum, 2)
            self.assertEqual(r.pagelen, 2)
    
    def test_extra_slice(self):
        schema = fields.Schema(key=fields.ID(stored=True))
        ix = RamStorage().create_index(schema)
        w = ix.writer()
        for char in u"abcdefghijklmnopqrstuvwxyz":
            w.add_document(key=char)
        w.commit()
        
        with ix.searcher() as s:
            r = s.search(query.Every(), limit=5)
            self.assertEqual(r[6:7], []) 
    
    def test_page_counts(self):
        from whoosh.scoring import Frequency
        
        schema = fields.Schema(id=fields.ID(stored=True))
        st = RamStorage()
        ix = st.create_index(schema)
        
        w = ix.writer()
        for i in xrange(10):
            w.add_document(id=unicode(i))
        w.commit()
        
        with ix.searcher(weighting=Frequency) as s:
            q = query.Every("id")
            
            r = s.search(q)
            self.assertEqual(len(r), 10)
            
            self.assertRaises(ValueError, s.search_page, q, 0)
            
            r = s.search_page(q, 1, 5)
            self.assertEqual(len(r), 10)
            self.assertEqual(r.pagecount, 2)
            
            r = s.search_page(q, 1, 5)
            self.assertEqual(len(r), 10)
            self.assertEqual(r.pagecount, 2)
            
            r = s.search_page(q, 2, 5)
            self.assertEqual(len(r), 10)
            self.assertEqual(r.pagecount, 2)
            self.assertEqual(r.pagenum, 2)
            
            r = s.search_page(q, 1, 10)
            self.assertEqual(len(r), 10)
            self.assertEqual(r.pagecount, 1)
            self.assertEqual(r.pagenum, 1)
    
    def test_resultspage(self):
        schema = fields.Schema(id=fields.STORED, content=fields.TEXT)
        ix = RamStorage().create_index(schema)
        
        domain = ("alfa", "bravo", "bravo", "charlie", "delta")
        w = ix.writer()
        for i, lst in enumerate(permutations(domain, 3)):
            w.add_document(id=unicode(i), content=u" ".join(lst))
        w.commit()
        
        with ix.searcher() as s:
            q = query.Term("content", u"bravo")
            r = s.search(q, limit=10)
            tops = list(r)
            
            rp = s.search_page(q, 1, pagelen=5)
            self.assertEqual(list(rp), tops[0:5])
            self.assertEqual(rp[10:], [])
            
            rp = s.search_page(q, 2, pagelen=5)
            self.assertEqual(list(rp), tops[5:10])
            
            rp = s.search_page(q, 1, pagelen=10)
            self.assertEqual(len(rp), 54)
            self.assertEqual(rp.pagecount, 6)
            rp = s.search_page(q, 6, pagelen=10)
            self.assertEqual(len(list(rp)), 4)
            self.assertTrue(rp.is_last_page())
            
            self.assertRaises(ValueError, s.search_page, q, 0)
            self.assertRaises(ValueError, s.search_page, q, 7)
            
            rp = s.search_page(query.Term("content", "glonk"), 1)
            self.assertEqual(len(rp), 0)
            self.assertTrue(rp.is_last_page())
    
    def test_snippets(self):
        from whoosh import highlight
        
        ana = analysis.StemmingAnalyzer()
        schema = fields.Schema(text=fields.TEXT(stored=True, analyzer=ana))
        ix = RamStorage().create_index(schema)
        w = ix.writer()
        w.add_document(text=u"Lay out the rough animation by creating the important poses where they occur on the timeline.")
        w.add_document(text=u"Set key frames on everything that's key-able. This is for control and predictability: you don't want to accidentally leave something un-keyed. This is also much faster than selecting the parameters to key.")
        w.add_document(text=u"Use constant (straight) or sometimes linear transitions between keyframes in the channel editor. This makes the character jump between poses.")
        w.add_document(text=u"Keying everything gives quick, immediate results, but it can become difficult to tweak the animation later, especially for complex characters.")
        w.add_document(text=u"Copy the current pose to create the next one: pose the character, key everything, then copy the keyframe in the playbar to another frame, and key everything at that frame.")
        w.commit()
        
        target = ["Set KEY frames on everything that's KEY-able. This is for control and predictability...leave something un-KEYED. This is also much faster than selecting...parameters to KEY",
                  "next one: pose the character, KEY everything, then copy...playbar to another frame, and KEY everything at that frame",
                  "KEYING everything gives quick, immediate results"]
        
        with ix.searcher() as s:
            qp = qparser.QueryParser("text", ix.schema)
            q = qp.parse(u"key")
            r = s.search(q)
            r.formatter = highlight.UppercaseFormatter()
            
            self.assertEqual([hit.highlights("text") for hit in r], target)
            
    def test_keyterms(self):
        ana = analysis.StandardAnalyzer()
        vectorformat = formats.Frequency(ana)
        schema = fields.Schema(path=fields.ID,
                               content=fields.TEXT(analyzer=ana,
                                                   vector=vectorformat))
        st = RamStorage()
        ix = st.create_index(schema)
        w = ix.writer()
        w.add_document(path=u"a",content=u"This is some generic content")
        w.add_document(path=u"b",content=u"This is some distinctive content")
        w.commit()
        
        with ix.searcher() as s:
            docnum = s.document_number(path=u"b")
            keyterms = list(s.key_terms([docnum], "content"))
            self.assertTrue(len(keyterms) > 0)
            self.assertEqual(keyterms[0][0], "distinctive")
            
            r = s.search(query.Term("path", u"b"))
            keyterms2 = list(r.key_terms("content"))
            self.assertTrue(len(keyterms2) > 0)
            self.assertEqual(keyterms2[0][0], "distinctive")
        
    def test_lengths(self):
        schema = fields.Schema(id=fields.STORED, text=fields.TEXT)
        ix = RamStorage().create_index(schema)
        
        w = ix.writer()
        w.add_document(id=1, text=u"alfa bravo charlie delta echo")
        w.add_document(id=2, text=u"bravo charlie delta echo foxtrot")
        w.add_document(id=3, text=u"charlie needle echo foxtrot golf")
        w.add_document(id=4, text=u"delta echo foxtrot golf hotel")
        w.add_document(id=5, text=u"echo needle needle hotel india")
        w.add_document(id=6, text=u"foxtrot golf hotel india juliet")
        w.add_document(id=7, text=u"golf needle india juliet kilo")
        w.add_document(id=8, text=u"hotel india juliet needle lima")
        w.commit()
        
        with ix.searcher() as s:
            q = query.Or([query.Term("text", u"needle"), query.Term("text", u"charlie")])
            r = s.search(q, limit=2)
            self.assertEqual(r.has_exact_length(), False)
            self.assertEqual(r.estimated_length(), 7)
            self.assertEqual(r.estimated_min_length(), 3)
            self.assertEqual(r.scored_length(), 2)
            self.assertEqual(len(r), 6)
    
    def test_lengths2(self):
        schema = fields.Schema(text=fields.TEXT(stored=True))
        ix = RamStorage().create_index(schema)
        count = 0
        for _ in xrange(3):
            w = ix.writer()
            for ls in permutations(u"alfa bravo charlie".split()):
                if "bravo" in ls and "charlie" in ls:
                    count += 1
                w.add_document(text=u" ".join(ls))
            w.commit(merge=False)
        
        with ix.searcher() as s:
            q = query.Or([query.Term("text", u"bravo"), query.Term("text", u"charlie")])
            r = s.search(q, limit=None)
            self.assertEqual(len(r), count)
            
            r = s.search(q, limit=3)
            self.assertEqual(len(r), count)


















if __name__ == '__main__':
    unittest.main()




        
    


import unittest

from whoosh import fields, scoring
from whoosh.filedb.filestore import RamStorage
from whoosh.qparser import QueryParser
from whoosh.query import *
from whoosh.spans import *


class TestQueries(unittest.TestCase):
    def test_all_terms(self):
        q = QueryParser("a", None).parse(u'hello b:there c:"my friend"')
        ts = set()
        q.all_terms(ts, phrases=False)
        self.assertEqual(sorted(ts), [("a", "hello"), ("b", "there")])
        ts = set()
        q.all_terms(ts, phrases=True)
        self.assertEqual(sorted(ts), [("a", "hello"), ("b", "there"),
                                      ("c", "friend"), ("c", "my")])
    
    def test_existing_terms(self):
        s = fields.Schema(key=fields.ID, value=fields.TEXT)
        st = RamStorage()
        ix = st.create_index(s)
        
        w = ix.writer()
        w.add_document(key=u"a", value=u"alfa bravo charlie delta echo")
        w.add_document(key=u"b", value=u"foxtrot golf hotel india juliet")
        w.commit()
        
        r = ix.reader()
        q = QueryParser("value", None).parse(u'alfa hotel tango "sierra bravo"')
        
        ts = q.existing_terms(r, phrases=False)
        self.assertEqual(sorted(ts), [("value", "alfa"), ("value", "hotel")])
        
        ts = q.existing_terms(r)
        self.assertEqual(sorted(ts), [("value", "alfa"), ("value", "bravo"), ("value", "hotel")])
        
        ts = set()
        q.existing_terms(r, ts, reverse=True)
        self.assertEqual(sorted(ts), [("value", "sierra"), ("value", "tango")])
    
    def test_replace(self):
        q = And([Or([Term("a", "b"), Term("b", "c")], boost=1.2), Variations("a", "b", boost=2.0)])
        q = q.replace("b", "BB")
        self.assertEqual(q, And([Or([Term("a", "BB"), Term("b", "c")], boost=1.2), Variations("a", "BB", boost=2.0)]))
    
    def test_apply(self):
        def visit(q):
            if isinstance(q, (Term, Variations, FuzzyTerm)):
                q.text = q.text.upper()
                return q
            return q.apply(visit)
        
        before = And([Not(Term("a", u"b")), Variations("a", u"c"), Not(FuzzyTerm("a", u"d"))])
        after = visit(before)
        self.assertEqual(after, And([Not(Term("a", u"B")), Variations("a", u"C"), Not(FuzzyTerm("a", u"D"))]))
        
        def term2var(q):
            if isinstance(q, Term):
                return Variations(q.fieldname, q.text)
            else:
                return q.apply(term2var)
    
        q = And([Term("f", "alfa"), Or([Term("f", "bravo"), Not(Term("f", "charlie"))])])
        q = term2var(q)
        self.assertEqual(q, And([Variations('f', 'alfa'), Or([Variations('f', 'bravo'), Not(Variations('f', 'charlie'))])]))
    
    def test_accept(self):
        def boost_phrases(q):
            if isinstance(q, Phrase):
                q.boost *= 2.0
            return q
        
        before = And([Term("a", u"b"), Or([Term("c", u"d"), Phrase("a", [u"e", u"f"])]), Phrase("a", [u"g", u"h"], boost=0.25)])
        after = before.accept(boost_phrases)
        self.assertEqual(after, And([Term("a", u"b"), Or([Term("c", u"d"), Phrase("a", [u"e", u"f"], boost=2.0)]), Phrase("a", [u"g", u"h"], boost=0.5)]))
    
        before = Phrase("a", [u"b", u"c"], boost=2.5)
        after = before.accept(boost_phrases)
        self.assertEqual(after, Phrase("a", [u"b", u"c"], boost=5.0))
    
    def test_simplify(self):
        s = fields.Schema(k=fields.ID, v=fields.TEXT)
        ix = RamStorage().create_index(s)
        
        w = ix.writer()
        w.add_document(k=u"1", v=u"aardvark apple allan alfa bear bee")
        w.add_document(k=u"2", v=u"brie glue geewhiz goop julia")
        w.commit()
        
        r = ix.reader()
        q1 = And([Prefix("v", "b", boost=2.0), Term("v", "juliet")])
        q2 = And([Or([Term('v', u'bear', boost=2.0), Term('v', u'bee', boost=2.0),
                      Term('v', u'brie', boost=2.0)]), Term('v', 'juliet')])
        self.assertEqual(q1.simplify(r), q2)
        
    def test_merge_ranges(self):
        q = And([TermRange("f1", u"a", None), TermRange("f1", None, u"z")])
        self.assertEqual(q.normalize(), TermRange("f1", u"a", u"z"))
        
        q = And([NumericRange("f1", None, u"aaaaa"), NumericRange("f1", u"zzzzz", None)])
        self.assertEqual(q.normalize(), q)
        
        q = And([TermRange("f1", u"a", u"z"), TermRange("f1", "b", "x")])
        self.assertEqual(q.normalize(), TermRange("f1", u"a", u"z"))
        
        q = And([TermRange("f1", u"a", u"m"), TermRange("f1", u"f", u"q")])
        self.assertEqual(q.normalize(), TermRange("f1", u"f", u"m"))
        
        q = Or([TermRange("f1", u"a", u"m"), TermRange("f1", u"f", u"q")])
        self.assertEqual(q.normalize(), TermRange("f1", u"a", u"q"))
        
        q = Or([TermRange("f1", u"m", None), TermRange("f1", None, u"n")])
        self.assertEqual(q.normalize(), Every("f1"))
        
        q = And([Every("f1"), Term("f1", "a"), Variations("f1", "b")])
        self.assertEqual(q.normalize(), Every("f1"))
        
        q = Or([Term("f1", u"q"), TermRange("f1", u"m", None), TermRange("f1", None, u"n")])
        self.assertEqual(q.normalize(), Every("f1"))
        
        q = And([Or([Term("f1", u"a"), Term("f1", u"b")]), Every("f1")])
        self.assertEqual(q.normalize(), Every("f1"))
        
        q = And([Term("f1", u"a"), And([Or([Every("f1")])])])
        self.assertEqual(q.normalize(), Every("f1"))
        
    def test_normalize_compound(self):
        def oq():
            return Or([Term("a", u"a"), Term("a", u"b")])
        def nq(level):
            if level == 0:
                return oq()
            else:
                return Or([nq(level-1), nq(level-1), nq(level-1)])
        
        q = nq(7)
        q = q.normalize()
        self.assertEqual(q, Or([Term("a", u"a"), Term("a", u"b")]))
    
    def test_duplicates(self):
        q = And([Term("a", u"b"), Term("a", u"b")])
        self.assertEqual(q.normalize(), Term("a", u"b"))
        
        q = And([Prefix("a", u"b"), Prefix("a", u"b")])
        self.assertEqual(q.normalize(), Prefix("a", u"b"))
        
        q = And([Variations("a", u"b"), And([Variations("a", u"b"), Term("a", u"b")])])
        self.assertEqual(q.normalize(), And([Variations("a", u"b"), Term("a", u"b")]))
        
        q = And([Term("a", u"b"), Prefix("a", u"b"), Term("a", u"b", boost=1.1)])
        self.assertEqual(q.normalize(), q)
        
        # Wildcard without * or ? normalizes to Term
        q = And([Wildcard("a", u"b"), And([Wildcard("a", u"b"), Term("a", u"b")])])
        self.assertEqual(q.normalize(), Term("a", u"b"))
    
    def test_query_copy_hash(self):
        def do(q1, q2):
            q1a = q1.copy()
            self.assertEqual(q1, q1a)
            self.assertEqual(hash(q1), hash(q1a))
            self.assertNotEqual(q1, q2)
            
        do(Term("a", u"b", boost=1.1), Term("a", u"b", boost=1.5))
        do(And([Term("a", u"b"), Term("c", u"d")], boost=1.1),
           And([Term("a", u"b"), Term("c", u"d")], boost=1.5))
        do(Or([Term("a", u"b", boost=1.1), Term("c", u"d")]),
           Or([Term("a", u"b", boost=1.8), Term("c", u"d")], boost=1.5))
        do(DisjunctionMax([Term("a", u"b", boost=1.8), Term("c", u"d")]),
           DisjunctionMax([Term("a", u"b", boost=1.1), Term("c", u"d")], boost=1.5))
        do(Not(Term("a", u"b", boost=1.1)), Not(Term("a", u"b", boost=1.5)))
        do(Prefix("a", u"b", boost=1.1), Prefix("a", u"b", boost=1.5))
        do(Wildcard("a", u"b*x?", boost=1.1), Wildcard("a", u"b*x?", boost=1.5))
        do(FuzzyTerm("a", u"b", constantscore=True),
           FuzzyTerm("a", u"b", constantscore=False))
        do(FuzzyTerm("a", u"b", boost=1.1), FuzzyTerm("a", u"b", boost=1.5))
        do(TermRange("a", u"b", u"c"), TermRange("a", u"b", u"d"))
        do(TermRange("a", None, u"c"), TermRange("a", None, None))
        do(TermRange("a", u"b", u"c", boost=1.1),
           TermRange("a", u"b", u"c", boost=1.5))
        do(TermRange("a", u"b", u"c", constantscore=True),
           TermRange("a", u"b", u"c", constantscore=False))
        do(NumericRange("a", 1, 5), NumericRange("a", 1, 6))
        do(NumericRange("a", None, 5), NumericRange("a", None, None))
        do(NumericRange("a", 3, 6, boost=1.1), NumericRange("a", 3, 6, boost=1.5))
        do(NumericRange("a", 3, 6, constantscore=True),
           NumericRange("a", 3, 6, constantscore=False))
        # do(DateRange)
        do(Variations("a", u"render"), Variations("a", u"renders"))
        do(Variations("a", u"render", boost=1.1),
           Variations("a", u"renders", boost=1.5))
        do(Phrase("a", [u"b", u"c", u"d"]), Phrase("a", [u"b", u"c", u"e"]))
        do(Phrase("a", [u"b", u"c", u"d"], boost=1.1),
           Phrase("a", [u"b", u"c", u"d"], boost=1.5))
        do(Phrase("a", [u"b", u"c", u"d"], slop=1),
           Phrase("a", [u"b", u"c", u"d"], slop=2))
        # do(Ordered)
        do(Every(), Every("a"))
        do(Every("a"), Every("b"))
        do(Every("a", boost=1.1), Every("a", boost=1.5))
        do(NullQuery, Term("a", u"b"))
        do(ConstantScoreQuery(Term("a", u"b")), ConstantScoreQuery(Term("a", u"c")))
        do(ConstantScoreQuery(Term("a", u"b"), score=2.0),
           ConstantScoreQuery(Term("a", u"c"), score=2.1))
        do(WeightingQuery(Term("a", u"b"), scoring.Frequency()),
           WeightingQuery(Term("a", u"c"), scoring.Frequency()))
        do(Require(Term("a", u"b"), Term("c", u"d")),
           Require(Term("a", u"b", boost=1.1), Term("c", u"d")))
        # do(Require)
        # do(AndMaybe)
        # do(AndNot)
        # do(Otherwise)
        
        do(SpanFirst(Term("a", u"b"), limit=1), SpanFirst(Term("a", u"b"), limit=2))
        do(SpanNear(Term("a", u"b"), Term("c", u"d")),
           SpanNear(Term("a", u"b"), Term("c", u"e")))
        do(SpanNear(Term("a", u"b"), Term("c", u"d"), slop=1),
           SpanNear(Term("a", u"b"), Term("c", u"d"), slop=2))
        do(SpanNear(Term("a", u"b"), Term("c", u"d"), mindist=1),
           SpanNear(Term("a", u"b"), Term("c", u"d"), mindist=2))
        do(SpanNear(Term("a", u"b"), Term("c", u"d"), ordered=True),
           SpanNear(Term("a", u"b"), Term("c", u"d"), ordered=False))
        do(SpanNot(Term("a", u"b"), Term("a", u"c")),
           SpanNot(Term("a", u"b"), Term("a", u"d")))
        do(SpanOr([Term("a", u"b"), Term("a", u"c"), Term("a", u"d")]),
           SpanOr([Term("a", u"b"), Term("a", u"c"), Term("a", u"e")]))
        do(SpanContains(Term("a", u"b"), Term("a", u"c")),
           SpanContains(Term("a", u"b"), Term("a", u"d")))
        # do(SpanBefore)
        # do(SpanCondition)
        


if __name__ == '__main__':
    unittest.main()



#===============================================================================
# Copyright 2007 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import sys
from struct import Struct, calcsize

IS_LITTLE = sys.byteorder == "little"

_INT_SIZE = calcsize("!i")
_SHORT_SIZE = calcsize("!H")
_LONG_SIZE = calcsize("!Q")
_FLOAT_SIZE = calcsize("!f")

_sbyte_struct = Struct("!b")
_ushort_struct = Struct("!H")
_int_struct = Struct("!i")
_uint_struct = Struct("!I")
_long_struct = Struct("!q")
_float_struct = Struct("!f")

pack_sbyte = _sbyte_struct.pack
pack_ushort = _ushort_struct.pack
pack_int = _int_struct.pack
pack_uint = _uint_struct.pack
pack_long = _long_struct.pack
pack_float = _float_struct.pack

unpack_sbyte = _sbyte_struct.unpack
unpack_ushort = _ushort_struct.unpack
unpack_int = _int_struct.unpack
unpack_uint = _uint_struct.unpack
unpack_long = _long_struct.unpack
unpack_float = _float_struct.unpack




#===============================================================================
# Copyright 2009 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""This module defines writer and reader classes for a fast, immutable
on-disk key-value database format. The current format is based heavily on
D. J. Bernstein's CDB format (http://cr.yp.to/cdb.html).
"""

from array import array
from collections import defaultdict
from cPickle import loads, dumps
from struct import Struct

from whoosh.system import (_INT_SIZE, _LONG_SIZE, pack_ushort, pack_uint,
                           pack_long, unpack_ushort, unpack_uint, unpack_long)
from whoosh.util import byte_to_length, utf8encode, utf8decode


_4GB = 4 * 1024 * 1024 * 1024

#def cdb_hash(key):
#    h = 5381L
#    for c in key:
#        h = (h + (h << 5)) & 0xffffffffL ^ ord(c)
#    return h

_header_entry_struct = Struct("!qI")  # Position, number of slots
header_entry_size = _header_entry_struct.size
pack_header_entry = _header_entry_struct.pack
unpack_header_entry = _header_entry_struct.unpack

_lengths_struct = Struct("!II")  # Length of key, length of data
lengths_size = _lengths_struct.size
pack_lengths = _lengths_struct.pack
unpack_lengths = _lengths_struct.unpack

_pointer_struct = Struct("!qq")  # Hash value, position
pointer_size = _pointer_struct.size
pack_pointer = _pointer_struct.pack
unpack_pointer = _pointer_struct.unpack

HEADER_SIZE = 256 * header_entry_size

#def _hash(value):
#    return abs(hash(value))
_hash = hash


# Table classes

class HashWriter(object):
    def __init__(self, dbfile):
        self.dbfile = dbfile
        # Seek past the first 2048 bytes of the file... we'll come back here
        # to write the header later
        dbfile.seek(HEADER_SIZE)
        # Store the directory of hashed values
        self.hashes = defaultdict(list)

    def add_all(self, items):
        dbfile = self.dbfile
        hashes = self.hashes
        pos = dbfile.tell()
        write = dbfile.write

        for key, value in items:
            write(pack_lengths(len(key), len(value)))
            write(key)
            write(value)

            h = _hash(key)
            hashes[h & 255].append((h, pos))
            pos += lengths_size + len(key) + len(value)

    def add(self, key, value):
        self.add_all(((key, value),))

    def _write_hashes(self):
        dbfile = self.dbfile
        hashes = self.hashes
        directory = self.directory = []

        pos = dbfile.tell()
        for i in xrange(0, 256):
            entries = hashes[i]
            numslots = 2 * len(entries)
            directory.append((pos, numslots))

            null = (0, 0)
            hashtable = [null] * numslots
            for hashval, position in entries:
                n = (hashval >> 8) % numslots
                while hashtable[n] != null:
                    n = (n + 1) % numslots
                hashtable[n] = (hashval, position)

            write = dbfile.write
            for hashval, position in hashtable:
                write(pack_pointer(hashval, position))
                pos += pointer_size

        dbfile.flush()

    def _write_directory(self):
        dbfile = self.dbfile
        directory = self.directory

        dbfile.seek(0)
        for position, numslots in directory:
            dbfile.write(pack_header_entry(position, numslots))
        assert dbfile.tell() == HEADER_SIZE
        dbfile.flush()

    def close(self):
        self._write_hashes()
        self._write_directory()
        self.dbfile.close()


class HashReader(object):
    def __init__(self, dbfile):
        self.dbfile = dbfile
        self.map = dbfile.map
        self.end_of_data = dbfile.get_long(0)
        self.is_closed = False

    def close(self):
        if self.is_closed:
            raise Exception("Tried to close %r twice" % self)
        del self.map
        self.dbfile.close()
        self.is_closed = True

    def read(self, position, length):
        return self.map[position:position + length]

    def _ranges(self, pos=HEADER_SIZE):
        eod = self.end_of_data
        read = self.read
        while pos < eod:
            keylen, datalen = unpack_lengths(read(pos, lengths_size))
            keypos = pos + lengths_size
            datapos = pos + lengths_size + keylen
            pos = datapos + datalen
            yield (keypos, keylen, datapos, datalen)

    def __iter__(self):
        return self.items()

    def items(self):
        read = self.read
        for keypos, keylen, datapos, datalen in self._ranges():
            yield (read(keypos, keylen), read(datapos, datalen))

    def keys(self):
        read = self.read
        for keypos, keylen, _, _ in self._ranges():
            yield read(keypos, keylen)

    def values(self):
        read = self.read
        for _, _, datapos, datalen in self._ranges():
            yield read(datapos, datalen)

    def __getitem__(self, key):
        for data in self.all(key):
            return data
        raise KeyError(key)

    def get(self, key, default=None):
        for data in self.all(key):
            return data
        return default

    def all(self, key):
        read = self.read
        for datapos, datalen in self._get_ranges(key):
            yield read(datapos, datalen)

    def __contains__(self, key):
        for _ in self._get_ranges(key):
            return True
        return False

    def _hashtable_info(self, keyhash):
        # Return (directory_position, number_of_hash_entries)
        return unpack_header_entry(self.read((keyhash & 255) * header_entry_size,
                                             header_entry_size))

    def _key_position(self, key):
        keyhash = _hash(key)
        hpos, hslots = self._hashtable_info(keyhash)
        if not hslots:
            raise KeyError(key)
        slotpos = hpos + (((keyhash >> 8) % hslots) * header_entry_size)
        
        return self.dbfile.get_long(slotpos + _INT_SIZE)

    def _key_at(self, pos):
        keylen = self.dbfile.get_uint(pos)
        return self.read(pos + lengths_size, keylen)

    def _get_ranges(self, key):
        read = self.read
        keyhash = _hash(key)
        hpos, hslots = self._hashtable_info(keyhash)
        if not hslots:
            return

        slotpos = hpos + (((keyhash >> 8) % hslots) * pointer_size)
        for _ in xrange(hslots):
            slothash, pos = unpack_pointer(read(slotpos, pointer_size))
            if not pos:
                return

            slotpos += pointer_size
            # If we reach the end of the hashtable, wrap around
            if slotpos == hpos + (hslots * pointer_size):
                slotpos = hpos

            if slothash == keyhash:
                keylen, datalen = unpack_lengths(read(pos, lengths_size))
                if keylen == len(key):
                    if key == read(pos + lengths_size, keylen):
                        yield (pos + lengths_size + keylen, datalen)
                        
    def end_of_hashes(self):
        lastpos, lastnum = unpack_header_entry(self.read(255 * header_entry_size,
                                                         header_entry_size))
        return lastpos + lastnum * pointer_size


class OrderedHashWriter(HashWriter):
    def __init__(self, dbfile):
        HashWriter.__init__(self, dbfile)
        self.index = []
        self.lastkey = None

    def add_all(self, items):
        dbfile = self.dbfile
        hashes = self.hashes
        pos = dbfile.tell()
        write = dbfile.write

        index = self.index
        lk = self.lastkey

        for key, value in items:
            if key <= lk:
                raise ValueError("Keys must increase: %r .. %r" % (lk, key))
            lk = key

            index.append(pos)
            write(pack_lengths(len(key), len(value)))
            write(key)
            write(value)

            h = _hash(key)
            hashes[h & 255].append((h, pos))
            
            pos += lengths_size + len(key) + len(value)
        
        self.lastkey = lk

    def close(self):
        self._write_hashes()
        dbfile = self.dbfile
        
        dbfile.write_uint(len(self.index))
        for n in self.index:
            dbfile.write_long(n)
        
        self._write_directory()
        self.dbfile.close()


class OrderedHashReader(HashReader):
    def __init__(self, dbfile):
        HashReader.__init__(self, dbfile)
        dbfile.seek(self.end_of_hashes())
        self.length = dbfile.read_uint()
        self.indexbase = dbfile.tell()
    
    def _closest_key(self, key):
        dbfile = self.dbfile
        key_at = self._key_at
        indexbase = self.indexbase
        lo = 0
        hi = self.length
        while lo < hi:
            mid = (lo + hi) // 2
            midkey = key_at(dbfile.get_long(indexbase + mid * _LONG_SIZE))
            if midkey < key:
                lo = mid + 1
            else:
                hi = mid
        #i = max(0, mid - 1)
        if lo == self.length:
            return None
        return dbfile.get_long(indexbase + lo * _LONG_SIZE)
    
    def closest_key(self, key):
        pos = self._closest_key(key)
        if pos is None:
            return None
        return self._key_at(pos)

    def _ranges_from(self, key):
        #read = self.read
        pos = self._closest_key(key)
        if pos is None:
            return

        for x in self._ranges(pos=pos):
            yield x

    def items_from(self, key):
        read = self.read
        for keypos, keylen, datapos, datalen in self._ranges_from(key):
            yield (read(keypos, keylen), read(datapos, datalen))

    def keys_from(self, key):
        read = self.read
        for keypos, keylen, _, _ in self._ranges_from(key):
            yield read(keypos, keylen)

    def values_from(self, key):
        read = self.read
        for _, _, datapos, datalen in self._ranges_from(key):
            yield read(datapos, datalen)


class CodedHashWriter(HashWriter):
    # Abstract base class, subclass must implement keycoder and valuecoder
    
    def __init__(self, dbfile):
        sup = super(CodedHashWriter, self)
        sup.__init__(dbfile)

        self._add = sup.add
        
    def add(self, key, data):
        self._add(self.keycoder(key), self.valuecoder(data))
        

class CodedHashReader(HashReader):
    # Abstract base class, subclass must implement keycoder, keydecoder and
    # valuecoder
    
    def __init__(self, dbfile):
        sup = super(CodedHashReader, self)
        sup.__init__(dbfile)

        self._items = sup.items
        self._keys = sup.keys
        self._get = sup.get
        self._getitem = sup.__getitem__
        self._contains = sup.__contains__
        
    def __getitem__(self, key):
        k = self.keycoder(key)
        return self.valuedecoder(self._getitem(k))

    def __contains__(self, key):
        return self._contains(self.keycoder(key))

    def get(self, key, default=None):
        k = self.keycoder(key)
        return self.valuedecoder(self._get(k, default))

    def items(self):
        kd = self.keydecoder
        vd = self.valuedecoder
        for key, value in self._items():
            yield (kd(key), vd(value))

    def keys(self):
        kd = self.keydecoder
        for k in self._keys():
            yield kd(k)


class CodedOrderedWriter(OrderedHashWriter):
    # Abstract base class, subclasses must implement keycoder and valuecoder
    
    def __init__(self, dbfile):
        sup = super(CodedOrderedWriter, self)
        sup.__init__(dbfile)
        self._add = sup.add

    def add(self, key, data):
        self._add(self.keycoder(key), self.valuecoder(data))


class CodedOrderedReader(OrderedHashReader):
    # Abstract base class, subclasses must implement keycoder, keydecoder,
    # and valuedecoder
    
    def __init__(self, dbfile):
        sup = super(CodedOrderedReader, self)
        sup.__init__(dbfile)

        self._items = sup.items
        self._items_from = sup.items_from
        self._keys = sup.keys
        self._keys_from = sup.keys_from
        self._get = sup.get
        self._getitem = sup.__getitem__
        self._contains = sup.__contains__

    def __getitem__(self, key):
        k = self.keycoder(key)
        return self.valuedecoder(self._getitem(k))

    def __contains__(self, key):
        try:
            codedkey = self.keycoder(key)
        except KeyError:
            return False
        return self._contains(codedkey)

    def get(self, key, default=None):
        k = self.keycoder(key)
        return self.valuedecoder(self._get(k, default))

    def items(self):
        kd = self.keydecoder
        vd = self.valuedecoder
        for key, value in self._items():
            yield (kd(key), vd(value))

    def items_from(self, key):
        fromkey = self.keycoder(key)
        kd = self.keydecoder
        vd = self.valuedecoder
        for key, value in self._items_from(fromkey):
            yield (kd(key), vd(value))

    def keys(self):
        kd = self.keydecoder
        for k in self._keys():
            yield kd(k)

    def keys_from(self, key):
        kd = self.keydecoder
        for k in self._keys_from(self.keycoder(key)):
            yield kd(k)


class TermIndexWriter(CodedOrderedWriter):
    def __init__(self, dbfile):
        super(TermIndexWriter, self).__init__(dbfile)
        self.fieldcounter = 0
        self.fieldmap = {}
    
    def keycoder(self, key):
        # Encode term
        fieldmap = self.fieldmap
        fieldname, text = key
        
        if fieldname in fieldmap:
            fieldnum = fieldmap[fieldname]
        else:
            fieldnum = self.fieldcounter
            fieldmap[fieldname] = fieldnum
            self.fieldcounter += 1
        
        key = pack_ushort(fieldnum) + utf8encode(text)[0]
        return key
    
    def valuecoder(self, data):
        w, offset, df = data
        
        if w == 1 and df == 1:
            v = dumps((offset, ), -1)
        elif w == df:
            v = dumps((offset, df), -1)
        else:
            v = dumps((w, offset, df), -1)
            
        # Strip off protocol at start and stack return command at end
        return v[2:-1]
            
    def close(self):
        self._write_hashes()
        dbfile = self.dbfile
        
        dbfile.write_uint(len(self.index))
        for n in self.index:
            dbfile.write_long(n)
        dbfile.write_pickle(self.fieldmap)
        
        self._write_directory()
        self.dbfile.close()


class TermIndexReader(CodedOrderedReader):
    def __init__(self, dbfile):
        super(TermIndexReader, self).__init__(dbfile)
        
        dbfile.seek(self.indexbase + self.length * _LONG_SIZE)
        self.fieldmap = dbfile.read_pickle()
        self.names = [None] * len(self.fieldmap)
        for name, num in self.fieldmap.iteritems():
            self.names[num] = name
    
    def keycoder(self, key):
        fieldname, text = key
        fnum = self.fieldmap.get(fieldname, 65535)
        return pack_ushort(fnum) + utf8encode(text)[0]
        
    def keydecoder(self, v):
        return (self.names[unpack_ushort(v[:2])[0]], utf8decode(v[2:])[0])
    
    def valuedecoder(self, v):
        v = loads(v + ".")
        if len(v) == 1:
            return (1, v[0], 1)
        elif len(v) == 2:
            return (v[1], v[0], v[1])
        else:
            return v
    

# docnum, fieldnum
_vectorkey_struct = Struct("!IH")


class TermVectorWriter(TermIndexWriter):
    def keycoder(self, key):
        fieldmap = self.fieldmap
        docnum, fieldname = key
        
        if fieldname in fieldmap:
            fieldnum = fieldmap[fieldname]
        else:
            fieldnum = self.fieldcounter
            fieldmap[fieldname] = fieldnum
            self.fieldcounter += 1
        
        return _vectorkey_struct.pack(docnum, fieldnum)
    
    def valuecoder(self, offset):
        return pack_long(offset)
        

class TermVectorReader(TermIndexReader):
    def keycoder(self, key):
        return _vectorkey_struct.pack(key[0], self.fieldmap[key[1]])
        
    def keydecoder(self, v):
        docnum, fieldnum = _vectorkey_struct.unpack(v)
        return (docnum, self.names[fieldnum])
    
    def valuedecoder(self, v):
        return unpack_long(v)[0]
    

class LengthWriter(object):
    def __init__(self, dbfile, doccount, lengths=None):
        self.dbfile = dbfile
        self.doccount = doccount
        if lengths is not None:
            self.lengths = lengths
        else:
            self.lengths = {}
    
    def add_all(self, items):
        lengths = self.lengths
        for docnum, fieldname, byte in items:
            if byte:
                if fieldname not in lengths:
                    lengths[fieldname] = array("B", (0 for _ in xrange(self.doccount)))
                lengths[fieldname][docnum] = byte
    
    def add(self, docnum, fieldname, byte):
        lengths = self.lengths
        if byte:
            if fieldname not in lengths:
                lengths[fieldname] = array("B", (0 for _ in xrange(self.doccount)))
            lengths[fieldname][docnum] = byte
    
    def reader(self):
        return LengthReader(None, self.doccount, lengths=self.lengths)
    
    def close(self):
        self.dbfile.write_ushort(len(self.lengths))
        for fieldname, arry in self.lengths.iteritems():
            self.dbfile.write_string(fieldname)
            self.dbfile.write_array(arry)
        self.dbfile.close()
        

class LengthReader(object):
    def __init__(self, dbfile, doccount, lengths=None):
        self.doccount = doccount
        
        if lengths is not None:
            self.lengths = lengths
        else:
            self.lengths = {}
            count = dbfile.read_ushort()
            for _ in xrange(count):
                fieldname = dbfile.read_string()
                self.lengths[fieldname] = dbfile.read_array("B", self.doccount)
            dbfile.close()
    
    def __iter__(self):
        for fieldname in self.lengths.keys():
            for docnum, byte in enumerate(self.lengths[fieldname]):
                yield docnum, fieldname, byte
    
    def get(self, docnum, fieldname, default=0):
        lengths = self.lengths
        if fieldname not in lengths:
            return default
        byte = lengths[fieldname][docnum] or default
        return byte_to_length(byte)
        

_stored_pointer_struct = Struct("!qI")  # offset, length
stored_pointer_size = _stored_pointer_struct.size
pack_stored_pointer = _stored_pointer_struct.pack
unpack_stored_pointer = _stored_pointer_struct.unpack


class StoredFieldWriter(object):
    def __init__(self, dbfile, fieldnames):
        self.dbfile = dbfile
        self.length = 0
        self.directory = []
        
        self.dbfile.write_long(0)
        self.dbfile.write_uint(0)
        
        self.name_map = {}
        for i, name in enumerate(fieldnames):
            self.name_map[name] = i
    
    def append(self, values):
        f = self.dbfile
        
        name_map = self.name_map
        
        vlist = [None] * len(name_map)
        for k, v in values.iteritems():
            if k in name_map:
                vlist[name_map[k]] = v
            else:
                # For dynamic stored fields, put them at the end of the list
                # as a tuple of (fieldname, value)
                vlist.append((k, v))
                
        v = dumps(vlist, -1)[2:-1]
        self.length += 1
        self.directory.append(pack_stored_pointer(f.tell(), len(v)))
        f.write(v)
    
    def close(self):
        f = self.dbfile
        directory_pos = f.tell()
        f.write_pickle(self.name_map)
        for pair in self.directory:
            f.write(pair)
        f.flush()
        f.seek(0)
        f.write_long(directory_pos)
        f.write_uint(self.length)
        f.close()


class StoredFieldReader(object):
    def __init__(self, dbfile):
        self.dbfile = dbfile

        dbfile.seek(0)
        pos = dbfile.read_long()
        self.length = dbfile.read_uint()
        
        dbfile.seek(pos)
        name_map = dbfile.read_pickle()
        self.names = [None] * len(name_map)
        for name, pos in name_map.iteritems():
            self.names[pos] = name
        self.directory_offset = dbfile.tell()
        
    def close(self):
        self.dbfile.close()

    def __getitem__(self, num):
        if num > self.length - 1:
            raise IndexError("Tried to get document %s, file has %s" % (num, self.length))
        
        dbfile = self.dbfile
        start = self.directory_offset + num * stored_pointer_size
        dbfile.seek(start)
        ptr = dbfile.read(stored_pointer_size)
        if len(ptr) != stored_pointer_size:
            raise Exception("Error reading %r @%s %s < %s" % (dbfile, start, len(ptr), stored_pointer_size))
        position, length = unpack_stored_pointer(ptr)
        vlist = loads(dbfile.map[position:position + length] + ".")
        
        names = self.names
        # Recreate a dictionary by putting the field names and values back
        # together by position. We can't just use dict(zip(...)) because we
        # want to filter out the None values.
        values = dict((names[i], vlist[i]) for i in xrange(len(names))
                      if vlist[i] is not None)
        
        # Pull out an extra stored dynamic field values off the end of the list
        if len(vlist) > len(names):
            values.update(dict(vlist[len(names):]))
        
        return values


# Utility functions

def dump_hash(hashreader):
    dbfile = hashreader.dbfile
    read = hashreader.read
    eod = hashreader.end_of_data

    print "HEADER_SIZE=", HEADER_SIZE, "eod=", eod

    # Dump hashtables
    for bucketnum in xrange(0, 256):
        pos, numslots = unpack_header_entry(read(bucketnum * header_entry_size, header_entry_size))
        if numslots:
            print "Bucket %d: %d slots" % (bucketnum, numslots)

            dbfile.seek(pos)
            for _ in xrange(0, numslots):
                print "  %X : %d" % unpack_pointer(read(pos, pointer_size))
                pos += pointer_size
        else:
            print "Bucket %d empty: %s, %s" % (bucketnum, pos, numslots)

    # Dump keys and values
    print "-----"
    pos = HEADER_SIZE
    dbfile.seek(pos)
    while pos < eod:
        keylen, datalen = unpack_lengths(read(pos, lengths_size))
        keypos = pos + lengths_size
        datapos = pos + lengths_size + keylen
        key = read(keypos, keylen)
        data = read(datapos, datalen)
        print "%d +%d,%d:%r->%r" % (pos, keylen, datalen, key, data)
        pos = datapos + datalen


##
#
#class FixedHashWriter(HashWriter):
#    def __init__(self, dbfile, keysize, datasize):
#        self.dbfile = dbfile
#        dbfile.seek(HEADER_SIZE)
#        self.hashes = defaultdict(list)
#        self.keysize = keysize
#        self.datasize = datasize
#        self.recordsize = keysize + datasize
#
#    def add_all(self, items):
#        dbfile = self.dbfile
#        hashes = self.hashes
#        recordsize = self.recordsize
#        pos = dbfile.tell()
#        write = dbfile.write
#
#        for key, value in items:
#            write(key + value)
#
#            h = _hash(key)
#            hashes[h & 255].append((h, pos))
#            pos += recordsize
#
#
#class FixedHashReader(HashReader):
#    def __init__(self, dbfile, keysize, datasize):
#        self.dbfile = dbfile
#        self.keysize = keysize
#        self.datasize = datasize
#        self.recordsize = keysize + datasize
#        
#        self.map = dbfile.map
#        self.end_of_data = dbfile.get_uint(0)
#        self.is_closed = False
#
#    def read(self, position, length):
#        return self.map[position:position + length]
#
#    def _ranges(self, pos=HEADER_SIZE):
#        keysize = self.keysize
#        recordsize = self.recordsize
#        eod = self.end_of_data
#        while pos < eod:
#            yield (pos, pos + keysize)
#            pos += recordsize
#
#    def __iter__(self):
#        return self.items()
#
#    def __contains__(self, key):
#        for _ in self._get_data_poses(key):
#            return True
#        return False
#
#    def items(self):
#        keysize = self.keysize
#        datasize = self.datasize
#        read = self.read
#        for keypos, datapos in self._ranges():
#            yield (read(keypos, keysize), read(datapos, datasize))
#
#    def keys(self):
#        keysize = self.keysize
#        read = self.read
#        for keypos, _ in self._ranges():
#            yield read(keypos, keysize)
#
#    def values(self):
#        datasize = self.datasize
#        read = self.read
#        for _, datapos in self._ranges():
#            yield read(datapos, datasize)
#
#    def __getitem__(self, key):
#        for data in self.all(key):
#            return data
#        raise KeyError(key)
#
#    def get(self, key, default=None):
#        for data in self.all(key):
#            return data
#        return default
#
#    def all(self, key):
#        datasize = self.datasize
#        read = self.read
#        for datapos in self._get_data_poses(key):
#            yield read(datapos, datasize)
#
#    def _key_at(self, pos):
#        return self.read(pos, self.keysize)
#
#    def _get_ranges(self, key):
#        raise NotImplementedError
#
#    def _get_data_poses(self, key):
#        keysize = self.keysize
#        read = self.read
#        keyhash = _hash(key)
#        hpos, hslots = self._hashtable_info(keyhash)
#        if not hslots:
#            return
#
#        slotpos = hpos + (((keyhash >> 8) % hslots) * pointer_size)
#        for _ in xrange(hslots):
#            slothash, pos = unpack_pointer(read(slotpos, pointer_size))
#            if not pos:
#                return
#
#            slotpos += pointer_size
#            # If we reach the end of the hashtable, wrap around
#            if slotpos == hpos + (hslots * pointer_size):
#                slotpos = hpos
#
#            if slothash == keyhash:
#                if key == read(pos, keysize):
#                    yield pos + keysize





#===============================================================================
# Copyright 2008 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""
This module contains classes for scoring (and sorting) search results.
"""

from __future__ import division
from math import log, pi


# Base classes

class WeightingModel(object):
    """Abstract base class for scoring models. A WeightingModel object provides
    a method, ``scorer``, which returns an instance of
    :class:`whoosh.scoring.Scorer`.
    
    Basically, WeightingModel objects store the configuration information for
    the model (for example, the values of B and K1 in the BM25F model), and
    then creates a scorer instance based on additional run-time information
    (the searcher, the fieldname, and term text) to do the actual scoring.
    """
    
    use_final = False
    
    def idf(self, searcher, fieldname, text):
        """Returns the inverse document frequency of the given term.
        """
        
        n = searcher.doc_frequency(fieldname, text)
        return log((searcher.doc_count_all()) / (n + 1)) + 1
    
    def scorer(self, searcher, fieldname, text, qf=1):
        """Returns an instance of :class:`whoosh.scoring.Scorer` configured
        for the given searcher, fieldname, and term text.
        """
        
        raise NotImplementedError(self.__class__.__name__)
    
    def final(self, searcher, docnum, score):
        """Returns a final score for each document. You can use this method
        in subclasses to apply document-level adjustments to the score, for
        example using the value of stored field to influence the score
        (although that would be slow).
        
        WeightingModel sub-classes that use ``final()`` should have the
        attribute ``use_final`` set to ``True``.
        
        :param searcher: :class:`whoosh.searching.Searcher` for the index.
        :param docnum: the doc number of the document being scored.
        :param score: the document's accumulated term score.
        
        :rtype: float
        """

        return score


class BaseScorer(object):
    """Base class for "scorer" implementations. A scorer provides a method for
    scoring a document, and sometimes methods for rating the "quality" of a
    document and a matcher's current "block", to implement quality-based
    optimizations.
    
    Scorer objects are created by WeightingModel objects. Basically,
    WeightingModel objects store the configuration information for the model
    (for example, the values of B and K1 in the BM25F model), and then creates
    a scorer instance.
    """
    
    def supports_quality(self):
        """Returns True if this class supports quality optimizations.
        """
        
        return False
    
    def score(self, matcher):
        """Returns a score for the current document of the matcher.
        """
        
        raise NotImplementedError(self.__class__.__name__)
    
    def quality(self, matcher):
        """Returns an approximate quality rating for the current document of
        the matcher.
        """
        
        raise NotImplementedError(self.__class__.__name__)
    
    def block_quality(self, matcher):
        """Returns an approximate quality rating for the matcher's current
        block (whatever concept of block the matcher might use).
        """
        
        raise NotImplementedError(self.__class__.__name__)


class WOLScorer(BaseScorer):
    """A "middleware" abstract base class for scorers that use
    weight-over-length (WOL) -- that is, weight divided by field length -- as
    the approximate quality rating. This class requires a method
    ``dfl(docnum)`` which returns the length of the field in the given
    document.
    """
    
    def supports_quality(self):
        return True
    
    def quality(self, matcher):
        return matcher.weight() / self.dfl(matcher.id())
    
    def block_quality(self, matcher):
        return matcher.block_maxwol()


# WeightScorer

class WeightScorer(BaseScorer):
    """A scorer that simply returns the weight as the score. This is useful
    for more complex weighting models to return when they are asked for a 
    scorer for fields that aren't scorable (don't store field lengths).
    """
    
    def supports_quality(self):
        return True
    
    def score(self, matcher):
        return matcher.weight()
    
    def quality(self, matcher):
        return matcher.weight()
    
    def block_quality(self, matcher):
        return matcher.block_maxweight()


# WeightingModel implementations

class BM25F(WeightingModel):
    """Implements the BM25F scoring algorithm.
    """
    
    def __init__(self, B=0.75, K1=1.2, **kwargs):
        """
        
        >>> from whoosh import scoring
        >>> # Set a custom B value for the "content" field
        >>> w = scoring.BM25F(B=0.75, content_B=1.0, K1=1.5)
        
        :param B: free parameter, see the BM25 literature. Keyword arguments of
            the form ``fieldname_B`` (for example, ``body_B``) set field-
            specific values for B.
        :param K1: free parameter, see the BM25 literature.
        """
        
        self.B = B
        self.K1 = K1
        
        self._field_B = {}
        for k, v in kwargs.iteritems():
            if k.endswith("_B"):
                fieldname = k[:-2]
                self._field_B[fieldname] = v
    
    def scorer(self, searcher, fieldname, text, qf=1):
        if not searcher.schema[fieldname].scorable:
            return WeightScorer()
        
        idf = searcher.idf(fieldname, text)
        avglength = searcher.avg_field_length(fieldname) or 1
        
        def dfl(docnum):
            return searcher.doc_field_length(docnum, fieldname, 1)
        
        if fieldname in self._field_B:
            B = self._field_B[fieldname]
        else:
            B = self.B
        
        return BM25F.BM25FScorer(idf, avglength, dfl, B, self.K1, qf=qf)
    
    class BM25FScorer(WOLScorer):
        def __init__(self, idf, avglength, dfl, B, K1, qf=1):
            self.idf = idf
            self.avglength = avglength
            self.dfl = dfl
            self.B = B
            self.K1 = K1
            self.qf = qf
        
        def score(self, matcher):
            weight = matcher.weight()
            length = self.dfl(matcher.id())
            B = self.B
            
            w = weight / ((1 - B) + B * (length / self.avglength))
            return self.idf * (w / (self.K1 + w))


class PL2(WeightingModel):
    """Implements the PL2 scoring model from Terrier.
    
    See http://terrier.org/
    """
    
    rec_log2_of_e = 1.0 / log(2)
    
    def __init__(self, c=1.0):
        self.c = c
        
    def scorer(self, searcher, fieldname, text, qf=1):
        if not searcher.schema[fieldname].scorable:
            return WeightScorer()
        
        collfreq = searcher.frequency(fieldname, text)
        doccount = searcher.doc_count_all()
        avglength = searcher.avg_field_length(fieldname) or 1
        
        def dfl(docnum):
            return searcher.doc_field_length(docnum, fieldname, 1)
        
        return PL2.PL2Scorer(collfreq, doccount, avglength, dfl, self.c, qf=qf)
    
    class PL2Scorer(WOLScorer):
        def __init__(self, collfreq, doccount, avglength, dfl, c, qf=1):
            self.collfreq = collfreq
            self.doccount = doccount
            self.avglength = avglength
            self.dfl = dfl
            self.c = c
            self.qf = qf
            
        def score(self, matcher):
            weight = matcher.weight()
            length = self.dfl(matcher.id())
            rec_log2_of_e = PL2.rec_log2_of_e
            
            tf = weight * log(1.0 + (self.c * self.avglength) / length)
            norm = 1.0 / (weight + 1.0)
            f = self.collfreq / self.doccount
            return (norm * self.qf * (tf * log(1.0 / f, 2)
                                      + f * rec_log2_of_e
                                      + 0.5 * log(2 * pi * tf, 2)
                                      + tf * (log(tf, 2) - rec_log2_of_e)))


# Simple models

class Frequency(WeightingModel):
    def scorer(self, searcher, fieldname, text, qf=1):
        return WeightScorer()
    

class TF_IDF(WeightingModel):
    def scorer(self, searcher, fieldname, text, qf=1):
        idf = searcher.idf(fieldname, text)
        return TF_IDF.TF_IDFScorer(idf)
    
    class TF_IDFScorer(BaseScorer):
        def __init__(self, idf):
            self.idf = idf
        
        def supports_quality(self):
            return True
        
        def score(self, matcher):
            return matcher.weight() * self.idf
        
        def quality(self, matcher):
            return matcher.weight()
        
        def block_quality(self, matcher):
            return matcher.block_maxweight()


# Utility models

class Weighting(WeightingModel):
    """This class provides backwards-compatibility with the old weighting
    class architecture, so any existing custom scorers don't need to be
    rewritten.
    
    It may also be useful for quick experimentation since you only need to
    override the ``score()`` method to try a scoring algorithm, without having
    to create an inner Scorer class::
    
        class MyWeighting(Weighting):
            def score(searcher, fieldname, text, docnum, weight):
                # Return the docnum as the score, for some reason
                return docnum
                
        mysearcher = myindex.searcher(weighting=MyWeighting)
    """
    
    def scorer(self, searcher, fieldname, text, qf=1):
        return self.CompatibilityScorer(searcher, fieldname, text, self.score)
    
    def score(self, searcher, fieldname, text, docnum, weight):
        raise NotImplementedError
    
    class CompatibilityScorer(BaseScorer):
        def __init__(self, searcher, fieldname, text, scoremethod):
            self.searcher = searcher
            self.fieldname = fieldname
            self.text = text
            self.scoremethod = scoremethod
        
        def score(self, matcher):
            return self.scoremethod(self.searcher, self.fieldname, self.text,
                                    matcher.id(), matcher.weight())


class MultiWeighting(WeightingModel):
    """Chooses from multiple scoring algorithms based on the field.
    """
    
    def __init__(self, default, **weightings):
        """The only non-keyword argument specifies the default
        :class:`Weighting` instance to use. Keyword arguments specify
        Weighting instances for specific fields.
        
        For example, to use ``BM25`` for most fields, but ``Frequency`` for
        the ``id`` field and ``TF_IDF`` for the ``keys`` field::
        
            mw = MultiWeighting(BM25(), id=Frequency(), keys=TF_IDF())
        
        :param default: the Weighting instance to use for fields not
            specified in the keyword arguments.
        """

        self.default = default
        # Store weighting functions by field name
        self.weightings = weightings
        
    def scorer(self, searcher, fieldname, text, qf=1):
        w = self.weightings.get(fieldname, self.default)
        return w.scorer(searcher, fieldname, text, qf=qf)
    
    
class ReverseWeighting(WeightingModel):
    """Wraps a weighting object and subtracts the wrapped model's scores from
    0, essentially reversing the weighting model.
    """
    
    def __init__(self, weighting):
        self.weighting = weighting
        
    def scorer(self, searcher, fieldname, text, qf=1):
        subscorer = self.weighting.scorer(searcher, fieldname, text, qf=qf)
        return ReverseWeighting.ReverseScorer(subscorer)
    
    class ReverseScorer(BaseScorer):
        def __init__(self, subscorer):
            self.subscorer = subscorer
        
        def supports_quality(self):
            return self.subscorer.supports_quality()
        
        def score(self, matcher):
            return 0 - self.subscorer.score(matcher)
        
        def quality(self, matcher):
            return 0 - self.subscorer.quality(matcher)
        
        def block_quality(self, matcher):
            return 0 - self.subscorer.block_quality(matcher)
        

#class PositionWeighting(WeightingModel):
#    def __init__(self, reversed=False):
#        self.reversed = reversed
#        
#    def scorer(self, searcher, fieldname, text, qf=1):
#        return PositionWeighting.PositionScorer()
#    
#    class PositionScorer(BaseScorer):
#        def score(self, matcher):
#            p = min(span.pos for span in matcher.spans())
#            if self.reversed:
#                return p
#            else:
#                return 0 - p





#===============================================================================
# Copyright 2007 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from __future__ import with_statement
import threading
import time

from whoosh.store import LockError
from whoosh.util import synchronized


# Exceptions

class IndexingError(Exception):
    pass


# Base class

class IndexWriter(object):
    """High-level object for writing to an index.
    
    To get a writer for a particular index, call
    :meth:`~whoosh.index.Index.writer` on the Index object.
    
    >>> writer = my_index.writer()
    
    You can use this object as a context manager. If an exception is thrown
    from within the context it calls cancel(), otherwise it calls commit() when
    the context exits.
    """
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.cancel()
        else:
            self.commit()
    
    def add_field(self, fieldname, fieldtype, **kwargs):
        """Adds a field to the index's schema.
        
        :param fieldname: the name of the field to add.
        :param fieldtype: an instantiated :class:`whoosh.fields.FieldType`
            object.
        """
        
        self.schema.add(fieldname, fieldtype, **kwargs)
    
    def remove_field(self, fieldname, **kwargs):
        """Removes the named field from the index's schema. Depending on the
        backend implementation, this may or may not actually remove existing
        data for the field from the index. Optimizing the index should always
        clear out existing data for a removed field.
        """
        
        self.schema.remove(fieldname, **kwargs)
        
    def reader(self, **kwargs):
        """Returns a reader for the existing index.
        """
        
        raise NotImplementedError
    
    def searcher(self, **kwargs):
        from whoosh.searching import Searcher
        
        return Searcher(self.reader(), **kwargs)
    
    def delete_by_term(self, fieldname, text, searcher=None):
        """Deletes any documents containing "term" in the "fieldname" field.
        This is useful when you have an indexed field containing a unique ID
        (such as "pathname") for each document.
        
        :returns: the number of documents deleted.
        """
        
        from whoosh.query import Term
        
        q = Term(fieldname, text)
        return self.delete_by_query(q, searcher=searcher)
    
    def delete_by_query(self, q, searcher=None):
        """Deletes any documents matching a query object.
        
        :returns: the number of documents deleted.
        """
        
        if searcher:
            s = searcher
        else:
            s = self.searcher()
        
        try:
            count = 0
            for docnum in s.docs_for_query(q):
                if not self.is_deleted(docnum):
                    self.delete_document(docnum)
                    count += 1
        finally:
            if not searcher:
                s.close()
        
        return count
    
    def delete_document(self, docnum, delete=True):
        """Deletes a document by number.
        """
        raise NotImplementedError
    
    def add_document(self, **fields):
        """The keyword arguments map field names to the values to index/store.
        
        For fields that are both indexed and stored, you can specify an
        alternate value to store using a keyword argument in the form
        "_stored_<fieldname>". For example, if you have a field named "title"
        and you want to index the text "a b c" but store the text "e f g", use
        keyword arguments like this::
        
            writer.add_document(title=u"a b c", _stored_title=u"e f g")
        """
        raise NotImplementedError
    
    def _unique_fields(self, fields):
        # Check which of the supplied fields are unique
        unique_fields = [name for name, field in self.schema.items()
                         if name in fields and field.unique]
        if not unique_fields:
            raise IndexingError("None of the fields in %r"
                                " are unique" % fields.keys())
        return unique_fields
    
    def update_document(self, **fields):
        """The keyword arguments map field names to the values to index/store.
        
        Note that this method will only replace a *committed* document;
        currently it cannot replace documents you've added to the IndexWriter
        but haven't yet committed. For example, if you do this:
        
        >>> writer.update_document(unique_id=u"1", content=u"Replace me")
        >>> writer.update_document(unique_id=u"1", content=u"Replacement")
        
        ...this will add two documents with the same value of ``unique_id``,
        instead of the second document replacing the first.
        
        For fields that are both indexed and stored, you can specify an
        alternate value to store using a keyword argument in the form
        "_stored_<fieldname>". For example, if you have a field named "title"
        and you want to index the text "a b c" but store the text "e f g", use
        keyword arguments like this::
        
            writer.update_document(title=u"a b c", _stored_title=u"e f g")
        """
        
        # Delete the set of documents matching the unique terms
        unique_fields = self._unique_fields(fields)
        with self.searcher() as s:
            for docnum in s._find_unique([(name, fields[name])
                                          for name in unique_fields]):
                self.delete_document(docnum)
        
        # Add the given fields
        self.add_document(**fields)
    
    def commit(self):
        """Finishes writing and unlocks the index.
        """
        pass
        
    def cancel(self):
        """Cancels any documents/deletions added by this object
        and unlocks the index.
        """
        pass
    

class PostingWriter(object):
    def start(self, format):
        """Start a new set of postings for a new term. Implementations may
        raise an exception if this is called without a corresponding call to
        finish().
        """
        raise NotImplementedError
    
    def write(self, id, weight, valuestring):
        """Add a posting with the given ID and value.
        """
        raise NotImplementedError
    
    def finish(self):
        """Finish writing the postings for the current term. Implementations
        may raise an exception if this is called without a preceding call to
        start().
        """
        pass
    
    def close(self):
        """Finish writing all postings and close the underlying file.
        """
        pass


class AsyncWriter(threading.Thread, IndexWriter):
    """Convenience wrapper for a writer object that might fail due to locking
    (i.e. the ``filedb`` writer). This object will attempt once to obtain the
    underlying writer, and if it's successful, will simply pass method calls on
    to it.
    
    If this object *can't* obtain a writer immediately, it will *buffer*
    delete, add, and update method calls in memory until you call ``commit()``.
    At that point, this object will start running in a separate thread, trying
    to obtain the writer over and over, and once it obtains it, "replay" all
    the buffered method calls on it.
    
    In a typical scenario where you're adding a single or a few documents to
    the index as the result of a Web transaction, this lets you just create the
    writer, add, and commit, without having to worry about index locks,
    retries, etc.
    
    For example, to get an aynchronous writer, instead of this:
    
    >>> writer = myindex.writer(postlimitmb=128)
    
    Do this:
    
    >>> from whoosh.writing import AsyncWriter
    >>> writer = AsyncWriter(myindex, )
    """
    
    def __init__(self, index, delay=0.25, writerargs=None):
        """
        :param index: the :class:`whoosh.index.Index` to write to.
        :param delay: the delay (in seconds) between attempts to instantiate
            the actual writer.
        :param writerargs: an optional dictionary specifying keyword arguments
            to to be passed to the index's ``writer()`` method.
        """
        
        threading.Thread.__init__(self)
        self.running = False
        self.index = index
        self.writerargs = writerargs or {}
        self.delay = delay
        self.events = []
        try:
            self.writer = self.index.writer(**self.writerargs)
        except LockError:
            self.writer = None
    
    def reader(self):
        return self.index.reader()
    
    def searcher(self, **kwargs):
        from whoosh.searching import Searcher
        return Searcher(self.reader(), fromindex=self.index, **kwargs)
    
    def _record(self, method, args, kwargs):
        if self.writer:
            getattr(self.writer, method)(*args, **kwargs)
        else:
            self.events.append((method, args, kwargs))
    
    def run(self):
        self.running = True
        writer = self.writer
        while writer is None:
            try:
                writer = self.writerfn(**self.writerargs)
            except LockError:
                time.sleep(self.delay)
        for method, args, kwargs in self.events:
            getattr(writer, method)(*args, **kwargs)
        writer.commit(*self.commitargs, **self.commitkwargs)
    
    def delete_document(self, *args, **kwargs):
        self._record("delete_document", args, kwargs)
    
    def add_document(self, *args, **kwargs):
        self._record("add_document", args, kwargs)
        
    def update_document(self, *args, **kwargs):
        self._record("update_document", args, kwargs)
    
    def add_field(self, *args, **kwargs):
        self._record("add_field", args, kwargs)
        
    def remove_field(self, *args, **kwargs):
        self._record("remove_field", args, kwargs)
    
    def delete_by_term(self, *args, **kwargs):
        self._record("delete_by_term", args, kwargs)
    
    def commit(self, *args, **kwargs):
        if self.writer:
            self.writer.commit(*args, **kwargs)
        else:
            self.commitargs, self.commitkwargs = args, kwargs
            self.start()
    
    def cancel(self, *args, **kwargs):
        if self.writer:
            self.writer.cancel(*args, **kwargs)
    

class BufferedWriter(IndexWriter):
    """Convenience class that acts like a writer but buffers added documents to
    a :class:`~whoosh.ramindex.RamIndex` before dumping the buffered documents
    as a batch into the actual index.
    
    In scenarios where you are continuously adding single documents very
    rapidly (for example a web application where lots of users are adding
    content simultaneously), using a BufferedWriter is *much* faster than
    opening and committing a writer for each document you add.
    
    (This class may also be useful for batches of ``update_document`` calls. In
    a normal writer, ``update_document`` calls cannot update documents you've
    added *in that writer*. With ``BufferedWriter``, this will work.)
    
    If you're adding a batches of documents at a time, you can just use a
    regular writer -- you're already committing a "batch" of documents, so you
    don't need this class.
    
    To use this class, create it from your index and *keep it open*, sharing
    it between threads.
    
    >>> from whoosh.writing import BufferedWriter
    >>> writer = BufferedWriter(myindex, period=120, limit=100)
    
    You can control how often the ``BufferedWriter`` flushes the in-memory
    index to disk using the ``period`` and ``limit`` arguments. ``period`` is
    the maximum number of seconds between commits. ``limit`` is the maximum
    number of additions to buffer between commits.
    
    You can read/search the combination of the on-disk index and the buffered
    documents in memory by calling ``BufferedWriter.reader()`` or
    ``BufferedWriter.searcher()``. This allows quasi-real-time search, where
    documents are available for searching as soon as they are buffered in
    memory, before they are committed to disk.
    
    >>> searcher = writer.searcher()
    
    .. tip::
        By using a searcher from the shared writer, multiple *threads* can
        search the buffered documents. Of course, other *processes* will only
        see the documents that have been written to disk. If you want indexed
        documents to become available to other processes as soon as possible,
        you have to use a traditional writer instead of a ``BufferedWriter``.
    
    Calling ``commit()`` on the ``BufferedWriter`` manually commits any batched
    up changes. You can continue to make changes after calling ``commit()``,
    and you can call ``commit()`` multiple times.
    
    .. note::
        This object keeps an underlying writer open and stores documents in
        memory, so you must explicitly call the :meth:`~BufferedWriter.close()`
        method on this object before it goes out of scope to release the
        write lock and make sure any uncommitted changes are saved.
    """

    def __init__(self, index, period=60, limit=10, writerargs=None,
                 commitargs=None, tempixclass=None):
        """
        :param index: the :class:`whoosh.index.Index` to write to.
        :param period: the maximum amount of time (in seconds) between commits.
            Set this to ``0`` or ``None`` to not use a timer. Do not set this
            any lower than a few seconds.
        :param limit: the maximum number of documents to buffer before
            committing.
        :param writerargs: dictionary specifying keyword arguments to be passed
            to the index's ``writer()`` method when creating a writer.
        :param commitargs: dictionary specifying keyword arguments to be passed
            to the writer's ``commit()`` method when committing a writer.
        """
        
        self.index = index
        self.period = period
        self.limit = limit
        self.writerargs = writerargs or {}
        self.commitargs = commitargs or {}
        self._sync_lock = threading.RLock()
        self._write_lock = threading.Lock()
        
        if tempixclass is None:
            from whoosh.ramindex import RamIndex as tempixclass
        self.tempixclass = tempixclass
        
        self.writer = None
        self.base = self.index.doc_count_all()
        self.bufferedcount = 0
        self.commitcount = 0
        self.ramindex = self._create_ramindex()
        if self.period:
            self.timer = threading.Timer(self.period, self.commit)
    
    def __del__(self):
        if hasattr(self, "writer") and self.writer:
            if not self.writer.is_closed:
                self.writer.cancel()
            del self.writer
    
    def _create_ramindex(self):
        return self.tempixclass(self.index.schema)
    
    def _get_writer(self):
        if self.writer is None:
            self.writer = self.index.writer(**self.writerargs)
            self.schema = self.writer.schema
            self.base = self.index.doc_count_all()
            self.bufferedcount = 0
        return self.writer
    
    @synchronized
    def reader(self, **kwargs):
        from whoosh.reading import MultiReader
        
        writer = self._get_writer()
        ramreader = self.ramindex
        if self.index.is_empty():
            return ramreader
        else:
            reader = writer.reader(**kwargs)
            if reader.is_atomic():
                reader = MultiReader([reader, ramreader])
            else:
                reader.add_reader(ramreader)
            return reader
    
    def searcher(self, **kwargs):
        from whoosh.searching import Searcher
        
        return Searcher(self.reader(), fromindex=self.index, **kwargs)
    
    def close(self):
        self.commit(restart=False)
    
    def commit(self, restart=True):
        if self.period:
            self.timer.cancel()
        
        # Replace the RAM index
        with self._sync_lock:
            oldramindex = self.ramindex
            self.ramindex = self._create_ramindex()
        
        with self._write_lock:
            if self.bufferedcount:
                self._get_writer().add_reader(oldramindex.reader())
                
            if self.writer:
                self.writer.commit(**self.commitargs)
                self.writer = None
                self.commitcount += 1
        
            if restart:
                if self.period:
                    self.timer = threading.Timer(self.period, self.commit)
    
    def add_reader(self, reader):
        with self._write_lock:
            self._get_writer().add_reader(reader)
    
    def add_document(self, **fields):
        with self._sync_lock:
            self.ramindex.add_document(**fields)
            self.bufferedcount += 1
        if self.bufferedcount >= self.limit:
            self.commit()
    
    @synchronized
    def update_document(self, **fields):
        self._get_writer()
        super(BufferedWriter, self).update_document(**fields)
    
    @synchronized
    def delete_document(self, docnum, delete=True):
        if docnum < self.base:
            return self._get_writer().delete_document(docnum, delete=delete)
        else:
            return self.ramindex.delete_document(docnum - self.base, delete=delete)
        
    @synchronized
    def is_deleted(self, docnum):
        if docnum < self.base:
            return self.writer.is_deleted(docnum)
        else:
            return self.ramindex.is_deleted(docnum - self.base)

# Backwards compatibility with old name
BatchWriter = BufferedWriter






#===============================================================================
# Copyright 2008 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""Classes and functions for classifying and extracting information from
documents.
"""

from __future__ import division
from collections import defaultdict
from math import log, sqrt


# Expansion models

class ExpansionModel(object):
    def __init__(self, doc_count, field_length):
        self.N = doc_count
        self.collection_total = field_length
        self.mean_length = self.collection_total / self.N
    
    def normalizer(self, maxweight, top_total):
        raise NotImplementedError
    
    def score(self, weight_in_top, weight_in_collection, top_total):
        raise NotImplementedError


class Bo1Model(ExpansionModel):
    def normalizer(self, maxweight, top_total):
        f = maxweight / self.N
        return (maxweight * log((1.0 + f) / f) + log(1.0 + f)) / log(2.0)
    
    def score(self, weight_in_top, weight_in_collection, top_total):
        f = weight_in_collection / self.N
        return weight_in_top * log((1.0 + f) / f, 2) + log(1.0 + f, 2)

 
class Bo2Model(ExpansionModel):
    def normalizer(self, maxweight, top_total):
        f = maxweight * self.N / self.collection_total
        return (maxweight * log((1.0 + f) / f, 2) + log(1.0 + f, 2))
    
    def score(self, weight_in_top, weight_in_collection, top_total):
        f = weight_in_top * top_total / self.collection_total
        return weight_in_top * log((1.0 + f) / f, 2) + log(1.0 + f, 2)


class KLModel(ExpansionModel):
    def normalizer(self, maxweight, top_total):
        return maxweight * log(self.collection_total / top_total) / log(2.0) * top_total
    
    def score(self, weight_in_top, weight_in_collection, top_total):
        wit_over_tt = weight_in_top / top_total
        wic_over_ct = weight_in_collection / self.collection_total
        
        if wit_over_tt < wic_over_ct:
            return 0
        else:
            return wit_over_tt * log((wit_over_tt) / (weight_in_top / self.collection_total), 2)


class Expander(object):
    """Uses an ExpansionModel to expand the set of query terms based on the top
    N result documents.
    """
    
    def __init__(self, ixreader, fieldname, model=Bo1Model):
        """
        :param reader: A :class:whoosh.reading.IndexReader object.
        :param fieldname: The name of the field in which to search.
        :param model: (classify.ExpansionModel) The model to use for expanding
            the query terms. If you omit this parameter, the expander uses
            scoring.Bo1Model by default.
        """
        
        self.ixreader = ixreader
        self.fieldname = fieldname
        
        if type(model) is type:
            model = model(self.ixreader.doc_count_all(),
                          self.ixreader.field_length(fieldname))
        self.model = model
        
        # Cache the collection frequency of every term in this field. This
        # turns out to be much faster than reading each individual weight
        # from the term index as we add words.
        self.collection_freq = dict((word, freq) for word, _, freq
                                      in self.ixreader.iter_field(self.fieldname))
        
        # Maps words to their weight in the top N documents.
        self.topN_weight = defaultdict(float)
        
        # Total weight of all terms in the top N documents.
        self.top_total = 0
    
    def add(self, vector):
        """Adds forward-index information about one of the "top N" documents.
        
        :param vector: A series of (text, weight) tuples, such as is
            returned by Reader.vector_as("weight", docnum, fieldname).
        """
        
        total_weight = 0
        topN_weight = self.topN_weight
        
        for word, weight in vector:
            total_weight += weight
            topN_weight[word] += weight
            
        self.top_total += total_weight
    
    def add_document(self, docnum):
        if self.ixreader.has_vector(docnum, self.fieldname):
            self.add(self.ixreader.vector_as("weight", docnum, self.fieldname))
        elif self.ixreader.schema[self.fieldname].stored:
            self.add_text(self.ixreader.stored_fields(docnum).get(self.fieldname))
        else:
            raise Exception("Field %r in document %s is not vectored or stored" % (self.fieldname, docnum))
    
    def add_text(self, string):
        field = self.ixreader.schema[self.fieldname]
        self.add((text, weight) for text, freq, weight, value
                 in field.index(string))
    
    def expanded_terms(self, number, normalize=True):
        """Returns the N most important terms in the vectors added so far.
        
        :param number: The number of terms to return.
        :param normalize: Whether to normalize the weights.
        :*returns*: A list of ("term", weight) tuples.
        """
        
        model = self.model
        tlist = []
        maxweight = 0
        collection_freq = self.collection_freq
        
        for word, weight in self.topN_weight.iteritems():
            if word in collection_freq:
                score = model.score(weight, collection_freq[word], self.top_total)
                if score > maxweight:
                    maxweight = score
                tlist.append((score, word))
        
        if normalize:
            norm = model.normalizer(maxweight, self.top_total)
        else:
            norm = maxweight
        tlist = [(weight / norm, t) for weight, t in tlist]
        tlist.sort(reverse=True)
        
        return [(t, weight) for weight, t in tlist[:number]]


# Clustering

def median(nums):
    nums = sorted(nums)
    l = len(nums)
    if l % 2:  # Odd
        return nums[l // 2]
    else:
        return (nums[l // 2 - 1] + nums[l // 2]) / 2.0


def mean(nums):
    return sum(nums) / len(nums)


def minkowski_distance(x, y, p=2):
    assert(len(y) == len(x))
    s = sum(abs(x[i] - y[i]) ** p for i in xrange(len(x)))
    return s ** 1.0 / p
   

def list_to_matrix(ls, f, symmetric=False, diagonal=None):
    matrix = []
    for rownum, i1 in enumerate(ls):
        row = []
        for colnum, i2 in enumerate(ls):
            if diagonal is not None and rownum == colnum:
                # Cell on the diagonal
                row.append(diagonal)
            elif symmetric and colnum < rownum:
                # Matrix is symmetrical and we've already calculated this cell
                # on the other side of the diagonal.
                row.append(matrix[colnum][rownum])
            else:
                row.append(f(i1, i2))
        matrix.append(row)
    return matrix


def magnitude(v):
    return sqrt(sum(v[i] ** 2 for i in xrange(len(v))))
    

def dot_product(v1, v2):
    assert len(v1) == len(v2)
    return sum(v1[i] * v2[i] for i in xrange(len(v1)))


def centroid(points, method=median):
    return tuple(method([point[i] for point in points])
                 for i in xrange(len(points[0])))


class Cluster(object):
    def __init__(self, *items):
        self.items = list(items)
    
    def __repr__(self):
        return "<C %r>" % (self.items, )
    
    def __len__(self):
        return len(self.items)
    
    def __add__(self, cluster):
        return Cluster(self.items + cluster.items)
    
    def __iter__(self):
        return iter(self.items)
    
    def __getitem__(self, n):
        return self.items.__getitem__(n)
    
    def append(self, item):
        self.items.append(item)
        
    def remove(self, item):
        self.items.remove(item)
        
    def pop(self, i=None):
        return self.items.pop(i)
    
    def flatten(self):
        for item in self.items:
            if isinstance(item, Cluster):
                for i2 in item.flatten():
                    yield i2
            else:
                yield item
                
    def dump(self, tab=0):
        print "%s-" % (" " * tab, )
        for item in self.items:
            if isinstance(item, Cluster):
                item.dump(tab + 2)
            else:
                print "%s%r" % (" " * tab, item)
    

class HierarchicalClustering(object):
    def __init__(self, distance_fn, linkage="uclus"):
        self.distance = distance_fn
        if linkage == "uclus":
            self.linkage = self.uclus_dist
        if linkage == "average":
            self.linkage = self.average_linkage_dist
        if linkage == "complete":
            self.linkage = self.complete_linkage_dist
        if linkage == "single":
            self.linkage = self.single_linkage_dist
    
    def uclus_dist(self, x, y):
        distances = []
        for xi in x.flatten():
            for yi in y.flatten():
                distances.append(self.distance(xi, yi))
        return median(distances)
    
    def average_linkage_dist(self, x, y):
        distances = []
        for xi in x.flatten():
            for yi in y.flatten():
                distances.append(self.distance(xi, yi))
        return mean(distances)
        
    def complete_linkage_dist(self, x, y):
        maxdist = self.distance(x[0], y[0])
        for xi in x.flatten():
            for yi in y.flatten():
                maxdist = max(maxdist, self.distance(xi, yi))
        return maxdist
   
    def single_linkage_dist(self, x, y):
        mindist = self.distance(x[0], y[0])
        for xi in x.flatten():
            for yi in y.flatten():
                mindist = min(mindist, self.distance(xi, yi))
        return mindist

    def clusters(self, data):
        data = [Cluster(x) for x in data]
        linkage = self.linkage
        matrix = None
        sequence = 0
        while matrix is None or len(matrix) > 2:
            matrix = list_to_matrix(data, linkage, True, 0)
            lowrow, lowcol = None, None
            mindist = None
            for rownum, row in enumerate(matrix):
                for colnum, cell in enumerate(row):
                    if rownum != colnum and (cell < mindist or lowrow is None):
                        lowrow, lowcol = rownum, colnum
                        mindist = cell
            
            sequence += 1
            cluster = Cluster(data[lowrow], data[lowcol])
            
            data.remove(data[max(lowrow, lowcol)])
            data.remove(data[min(lowrow, lowcol)])
            data.append(cluster)
        
        if isinstance(data, list):
            data = Cluster(*data)
        return data


class KMeansClustering(object):
    def __init__(self, distance_fn=None):
        self.distance = distance_fn or minkowski_distance
        
    def clusters(self, data, count):
        if len(data) > 1 and isinstance(data[0], (list, tuple)):
            l = len(data[0])
            if not all(len(item) == l for item in data[1:]):
                raise ValueError("All items in %r are not of the same dimension" % (data, ))
        if count <= 1:
            raise ValueError("You must ask for at least 2 clusters")
        if not data or len(data) == 1 or count >= len(data):
            return data
        
        clusters = [Cluster() for _ in xrange(count)]
        for i, item in enumerate(data):
            clusters[i % count].append(item)
        
        def move_item(item, pos, origin):
            closest = origin
            for cluster in clusters:
                if (self.distance(item, centroid(cluster))
                    < self.distance(item, centroid(closest))):
                    closest = cluster
            if closest is not origin:
                closest.append(origin.pop(pos))
                return True
            return False
        
        moved = True
        while moved:
            moved = False
            for cluster in clusters:
                for pos, item in enumerate(cluster):
                    moved = move_item(item, pos, cluster) or moved
                    
        return clusters
                    
        
# Similarity functions

def shingles(input, size=2):
    d = defaultdict(int)
    for shingle in (input[i:i + size]
                    for i in xrange(len(input) - (size - 1))):
        d[shingle] += 1
    return d.iteritems()


def simhash(features, hashbits=32):
    if hashbits == 32:
        hashfn = hash
    else:
        hashfn = lambda s: _hash(s, hashbits)
    
    vs = [0] * hashbits
    for feature, weight in features:
        h = hashfn(feature)
        for i in xrange(hashbits):
            if h & (1 << i):
                vs[i] += weight
            else:
                vs[i] -= weight
    
    out = 0
    for i, v in enumerate(vs):
        if v > 0:
            out |= 1 << i
    return out


def _hash(s, hashbits):
    # A variable-length version of Python's builtin hash
    if s == "":
        return 0
    else:
        x = ord(s[0]) << 7
        m = 1000003
        mask = 2 ** hashbits - 1
        for c in s:
            x = ((x * m) ^ ord(c)) & mask
        x ^= len(s)
        if x == -1: 
            x = -2
        return x

    
def hamming_distance(first_hash, other_hash, hashbits=32):
    x = (first_hash ^ other_hash) & ((1 << hashbits) - 1)
    tot = 0
    while x:
        tot += 1
        x &= x - 1
    return tot









import shutil
import tempfile

from whoosh.filedb.filestore import FileStorage


class TempStorage(object):
    def __init__(self, basename='', parentdir=None, suppress=frozenset(),
                 keepdir=False):
        self.basename = basename
        self.parentdir = parentdir
        self.suppress = suppress
        self.keepdir = keepdir
        self.dir = None
    
    def __enter__(self):
        self.dir = tempfile.mkdtemp(prefix=self.basename, suffix=".tmpix",
                                    dir=self.parentdir)
        return FileStorage(self.dir)
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            if exc_type not in self.suppress:
                print "Temp dir=", self.dir
                return False
        
        if not self.keepdir:
            try:
                shutil.rmtree(self.dir)
            except OSError, e:
                print "Can't remove temp dir: " + str(e)


class TempIndex(TempStorage):
    def __init__(self, schema, ixname='', **kwargs):
        super(TempIndex, self).__init__(basename=ixname, **kwargs)
        self.schema = schema

    def __enter__(self):
        fstore = super(TempIndex, self).__enter__()
        return fstore.create_index(self.schema, indexname=self.basename)





import unittest
from datetime import datetime

from whoosh.qparser.dateparse import *


basedate = datetime(2010, 9, 20, 15, 16, 6, 454000)
english = English()


class TestDateParser(unittest.TestCase):
    def assert_adatetime(self, at, **kwargs):
        self.assertEqual(at.__class__, adatetime)
        for key in adatetime.units:
            val = getattr(at, key)
            target = kwargs.get(key)
            self.assertEqual(val, target, "at.%s=%r not %r in %r" % (key, val, target, kwargs))
    
    def assert_timespan(self, ts, sargs, eargs):
        self.assert_adatetime(ts.start, **sargs)
        self.assert_adatetime(ts.end, **eargs)
    
    def assert_unamb(self, ts, **kwargs):
        self.assert_unamb_span(ts, kwargs, kwargs)
    
    def assert_unamb_span(self, ts, sargs, eargs):
        startdt = adatetime(**sargs).floor()
        enddt = adatetime(**eargs).ceil()
        self.assertEqual(ts.start, startdt, "start %s != %s" % (ts.start, startdt))
        self.assertEqual(ts.end, enddt, "end %s != %s" % (ts.end, enddt))
    
    def assert_datespan(self, ts, startdate, enddate):
        self.assertEqual(ts.__class__, timespan)
        self.assertEqual(ts.start, startdate)
        self.assertEqual(ts.end, enddate)
    
    #
    
    def test_simple(self, t=english.simple):
        self.assert_adatetime(t.date_from("2005", basedate), year=2005)
        self.assert_adatetime(t.date_from("200505", basedate), year=2005, month=5)
        self.assert_adatetime(t.date_from("20050510", basedate), year=2005, month=5, day=10)
        self.assert_adatetime(t.date_from("2005051001", basedate), year=2005, month=5, day=10, hour=1)
        self.assert_adatetime(t.date_from("200505100108", basedate), year=2005, month=5, day=10, hour=1, minute=8)
        self.assert_adatetime(t.date_from("20050510010835", basedate), year=2005, month=5, day=10, hour=1, minute=8, second=35)
        
        self.assert_adatetime(t.date_from("2005-05", basedate), year=2005, month=5)
        self.assert_adatetime(t.date_from("2005 05 10", basedate), year=2005, month=5, day=10)
        self.assert_adatetime(t.date_from("2005.05.10.01", basedate), year=2005, month=5, day=10, hour=1)
        self.assert_adatetime(t.date_from("2005/05/10 01:08", basedate), year=2005, month=5, day=10, hour=1, minute=8)
        self.assert_adatetime(t.date_from("2005.05.10  01:08:35", basedate), year=2005, month=5, day=10, hour=1, minute=8, second=35)
        
        self.assertEqual(t.date_from("2005 02 31", basedate), None)
        self.assertEqual(t.date_from("2005-13-32", basedate), None)
        
    def test_time(self, t=english.time):
        self.assert_adatetime(t.date_from("13:05", basedate), hour=13, minute=5)
        self.assertEqual(t.date_from("28:91", basedate), None)
        
        self.assert_adatetime(t.date_from("3pm", basedate), hour=15)
        self.assert_adatetime(t.date_from("3 pm", basedate), hour=15)
        self.assert_adatetime(t.date_from("10pm", basedate), hour=22)
        self.assert_adatetime(t.date_from("10 pm", basedate), hour=22)
        self.assert_adatetime(t.date_from("3am", basedate), hour=3)
        self.assert_adatetime(t.date_from("3:15 am", basedate), hour=3, minute=15)
        self.assert_adatetime(t.date_from("5:10pm", basedate), hour=17, minute=10)
        self.assert_adatetime(t.date_from("12:45am", basedate), hour=0, minute=45)
        self.assert_adatetime(t.date_from("12:45pm", basedate), hour=12, minute=45)
        self.assert_adatetime(t.date_from("5:45:05 pm", basedate), hour=17, minute=45, second=5)
        
        self.assert_adatetime(t.date_from("noon", basedate), hour=12, minute=0, second=0, microsecond=0)
        self.assert_adatetime(t.date_from("midnight", basedate), hour=0, minute=0, second=0, microsecond=0)
        
        self.assertEqual(t.date_from("15 am", basedate), None)
        self.assertEqual(t.date_from("24:00", basedate), None)
        self.assertEqual(t.date_from("12:65", basedate), None)
    
    def test_dmy(self, d=english.dmy):
        self.assert_adatetime(d.date_from("25 may 2011", basedate), year=2011, month=5, day=25)
        self.assert_adatetime(d.date_from("may 2 2011", basedate), year=2011, month=5, day=2)
        self.assert_adatetime(d.date_from("2011 25 may", basedate), year=2011, month=5, day=25)
        self.assert_adatetime(d.date_from("2011 may 5", basedate), year=2011, month=5, day=5)
        
        self.assert_adatetime(d.date_from("apr", basedate), month=4)
        self.assert_adatetime(d.date_from("september", basedate), month=9)
        self.assert_adatetime(d.date_from("2001", basedate), year=2001)
        self.assert_adatetime(d.date_from("july 2525", basedate), year=2525, month=7)
        self.assert_adatetime(d.date_from("nov 30", basedate), month=11, day=30)
        self.assertEqual(d.date_from("25 2525", basedate), None)
        
        self.assert_adatetime(d.date_from("25 may, 2011", basedate), year=2011, month=5, day=25)
        self.assert_adatetime(d.date_from("may 2nd, 2011", basedate), year=2011, month=5, day=2)
        self.assert_adatetime(d.date_from("2011, 25 may", basedate), year=2011, month=5, day=25)
        self.assert_adatetime(d.date_from("2011, may 5th", basedate), year=2011, month=5, day=5)
        
        self.assert_adatetime(d.date_from("today", basedate), year=2010, month=9, day=20)
        self.assert_adatetime(d.date_from("tomorrow", basedate), year=2010, month=9, day=21)
        self.assert_adatetime(d.date_from("yesterday", basedate), year=2010, month=9, day=19)
        self.assert_adatetime(d.date_from("this month", basedate), year=2010, month=9)
        self.assert_adatetime(d.date_from("this year", basedate), year=2010)
        
        self.assertEqual(d.date_from("now", basedate), basedate)
        
    def test_plustime(self, rt=english.plusdate):
        self.assertEqual(rt.date_from("+1hr", basedate),
                         basedate + timedelta(hours=1))
        self.assertEqual(rt.date_from("+5mins", basedate),
                         basedate + timedelta(minutes=5))
        self.assertEqual(rt.date_from("+20s", basedate),
                         basedate + timedelta(seconds=20))
        
        self.assertEqual(rt.date_from("- 2 h", basedate),
                         basedate + timedelta(hours=-2))
        self.assertEqual(rt.date_from("- 25 minutes", basedate),
                         basedate + timedelta(minutes=-25))
        self.assertEqual(rt.date_from("-400 secs", basedate),
                         basedate + timedelta(seconds=-400))
        
        self.assertEqual(rt.date_from("+1hr 5m", basedate),
                         basedate + timedelta(hours=1, minutes=5))
        self.assertEqual(rt.date_from("-8hr 12m", basedate),
                         basedate + timedelta(hours=-8, minutes=-12))
        self.assertEqual(rt.date_from("+1hr 5s", basedate),
                         basedate + timedelta(hours=1, seconds=5))
        self.assertEqual(rt.date_from("+1hr 12m 5s", basedate),
                         basedate + timedelta(hours=1, minutes=12, seconds=5))
        self.assertEqual(rt.date_from("-1hr 5s", basedate),
                         basedate + timedelta(hours=-1, seconds=-5))
        self.assertEqual(rt.date_from("-1hr 12m 5s", basedate),
                         basedate + timedelta(hours=-1, minutes=-12, seconds=-5))
        
    def test_relative_days(self):
        # "next monday" on monday
        self.assertEqual(relative_days(0, 0, 1), 7)
        # "last monday" on monday
        self.assertEqual(relative_days(0, 0, -1), -7)
        # "next tuesday" on wednesday
        self.assertEqual(relative_days(2, 1, 1), 6)
        # "last tuesday" on wednesay
        self.assertEqual(relative_days(2, 1, -1), -1)
        # "last monday" on sunday
        self.assertEqual(relative_days(6, 0, -1), -6)
        # "next monday" on sunday
        self.assertEqual(relative_days(6, 0, 1), 1)
        # "next wednesday" on tuesday
        self.assertEqual(relative_days(1, 2, 1), 1)
        # "last wednesday" on tuesday
        self.assertEqual(relative_days(1, 2, -1), -6)
        # "last wednesday" on thursday
        self.assertEqual(relative_days(3, 2, -1), -1)
        # "next wednesday" on thursday
        self.assertEqual(relative_days(3, 2, 1), 6)
        # "last wednesday" on tuesday
        self.assertEqual(relative_days(1, 2, -1), -6)
        # "next wednesday" on tuesday
        self.assertEqual(relative_days(1, 2, 1), 1)
        
    def test_dayname(self, p=english.dayname):
        self.assert_adatetime(p.date_from("next tuesday", basedate), year=2010, month=9, day=21)
        self.assert_adatetime(p.date_from("last tuesday", basedate), year=2010, month=9, day=14)
        self.assert_adatetime(p.date_from("next sunday", basedate), year=2010, month=9, day=26)
        self.assert_adatetime(p.date_from("last sun", basedate), year=2010, month=9, day=19)
        self.assert_adatetime(p.date_from("next th", basedate), year=2010, month=9, day=23)
        
    def test_reldate(self, p=english.plusdate):
        self.assertEqual(p.date_from("+1y", basedate),
                         basedate + relativedelta(years=1))
        self.assertEqual(p.date_from("+2mo", basedate),
                         basedate + relativedelta(months=2))
        self.assertEqual(p.date_from("+3w", basedate),
                         basedate + relativedelta(weeks=3))
        self.assertEqual(p.date_from("+5d", basedate),
                         basedate + relativedelta(days=5))
        self.assertEqual(p.date_from("+5days", basedate),
                         basedate + relativedelta(days=5))
        
        self.assertEqual(p.date_from("-6yr", basedate),
                         basedate + relativedelta(years=-6))
        self.assertEqual(p.date_from("- 7 mons", basedate),
                         basedate + relativedelta(months=-7))
        self.assertEqual(p.date_from("-8 wks", basedate),
                         basedate + relativedelta(weeks=-8))
        self.assertEqual(p.date_from("- 9 dy", basedate),
                         basedate + relativedelta(days=-9))
        
        
        self.assertEqual(p.date_from("+1y 12mo 400d", basedate),
                         basedate + relativedelta(years=1, months=12, days=400))
        self.assertEqual(p.date_from("-7mo 8d", basedate),
                         basedate + relativedelta(months=-7, days=-8))
        self.assertEqual(p.date_from("+5wks 2d", basedate),
                         basedate + relativedelta(weeks=5, days=2))
        self.assertEqual(p.date_from("-1y 1w", basedate),
                         basedate + relativedelta(years=-1, weeks=-1))
        
        self.assertEqual(p.date_from("+1y 2d 5h 12s", basedate),
                         basedate + relativedelta(years=1, days=2, hours=5, seconds=12))
        
    def test_bundle_subs(self, p=english.bundle):
        self.test_time(p)
        self.test_dmy(p)
        self.test_plustime(p)
        self.test_dayname(p)
        self.test_reldate(p)
        
    def test_bundle(self, p=english.bundle):
        self.assert_adatetime(p.date_from("mar 29 1972 2:45am", basedate),
                          year=1972, month=3, day=29, hour=2, minute=45)
        self.assert_adatetime(p.date_from("16:10:45 14 February 2005", basedate),
                          year=2005, month=2, day=14, hour=16, minute=10, second=45)
        self.assert_adatetime(p.date_from("1985 sept 12 12:01", basedate),
                          year=1985, month=9, day=12, hour=12, minute=1)
        self.assert_adatetime(p.date_from("5pm 21st oct 2005", basedate),
                          year=2005, month=10, day=21, hour=17)
        self.assert_adatetime(p.date_from("5:59:59pm next thur", basedate),
                          year=2010, month=9, day=23, hour=17, minute=59, second=59)
    
    def test_ranges(self, p=english.torange):
        self.assert_timespan(p.date_from("last tuesday to next tuesday", basedate),
                             dict(year=2010, month=9, day=14),
                             dict(year=2010, month=9, day=21))
        self.assert_timespan(p.date_from("last monday to dec 25", basedate),
                             dict(year=2010, month=9, day=13),
                             dict(year=None, month=12, day=25))
        self.assert_timespan(p.date_from("oct 25 to feb 14", basedate),
                             dict(year=None, month=10, day=25),
                             dict(year=None, month=2, day=14))
        self.assert_timespan(p.date_from("3am oct 12 to 5pm", basedate),
                             dict(year=None, month=10, day=12, hour=3),
                             dict(year=None, month=None, day=None, hour=17))
        self.assert_timespan(p.date_from("3am feb 12 to 5pm today", basedate),
                             dict(year=None, month=2, day=12, hour=3),
                             dict(year=2010, month=9, day=20, hour=17))
        self.assert_timespan(p.date_from("feb to oct", basedate),
                             dict(year=None, month=2),
                             dict(year=None, month=10))
        self.assert_timespan(p.date_from("oct 25 2005 11am to 5pm tomorrow", basedate),
                             dict(year=2005, month=10, day=25, hour=11),
                             dict(year=2010, month=9, day=21, hour=17))
        self.assert_timespan(p.date_from("oct 5 2005 to november 20", basedate),
                             dict(year=2005, month=10, day=5),
                             dict(year=None, month=11, day=20))
        self.assert_timespan(p.date_from("2007 to 2010", basedate),
                             dict(year=2007, month=None, day=None),
                             dict(year=2010, month=None, day=None))
        self.assert_timespan(p.date_from("2007 to oct 12", basedate),
                             dict(year=2007, month=None, day=None),
                             dict(year=None, month=10, day=12))
        
        self.assert_datespan(p.date_from("-2d to +1w", basedate),
                             basedate + relativedelta(days=-2),
                             basedate + relativedelta(weeks=1))
    
    def test_all(self):
        p = english.all
        self.test_bundle_subs(p)
        self.test_bundle(p)
        self.test_ranges(p)
    
    def test_final_dates(self, p=english):
        self.assert_unamb(p.date_from("5:10pm", basedate),
                          year=2010, month=9, day=20, hour=17, minute=10)
    
        self.assertEqual(p.date_from("may 32 2005", basedate), None)
        self.assertEqual(p.date_from("2005 may 32", basedate), None)            
        self.assertEqual(p.date_from("2005-13-32", basedate), None)
    
    def test_final_ranges(self, p=english):
        self.assert_unamb_span(p.date_from("feb to nov", basedate),
                               dict(year=2010, month=2),
                               dict(year=2010, month=11))
        
        # 2005 to 10 oct 2009 -> jan 1 2005 to oct 31 2009
        self.assert_unamb_span(p.date_from("2005 to 10 oct 2009", basedate),
                               dict(year=2005),
                               dict(year=2009, month=10, day=10))
        
        # jan 12 to oct 10 2009 -> jan 12 2009 to oct 10 2009
        self.assert_unamb_span(p.date_from("jan 12 to oct 10 2009", basedate),
                               dict(year=2009, month=1, day=12),
                               dict(year=2009, month=10, day=10))
        
        # jan to oct 2009 -> jan 1 2009 to oct 31 2009
        self.assert_unamb_span(p.date_from("jan to oct 2009", basedate),
                               dict(year=2009, month=1),
                               dict(year=2009, month=10, day=31))
        
        # mar 2005 to oct -> mar 1 2005 to oct 31 basedate.year
        self.assert_unamb_span(p.date_from("mar 2005 to oct", basedate),
                               dict(year=2005, month=3),
                               dict(year=2010, month=10, day=31))
        
        # jan 10 to jan 25 -> jan 10 basedate.year to jan 25 basedate.year
        self.assert_unamb_span(p.date_from("jan 10 to jan 25", basedate),
                               dict(year=2010, month=1, day=10),
                               dict(year=2010, month=1, day=25))
        
        # jan 2005 to feb 2009 -> jan 1 2005 to feb 28 2009
        self.assert_unamb_span(p.date_from("jan 2005 to feb 2009", basedate),
                               dict(year=2005, month=1),
                               dict(year=2009, month=2))
        
        # jan 5000 to mar -> jan 1 5000 to mar 5000
        self.assert_unamb_span(p.date_from("jan 5000 to mar", basedate),
                               dict(year=5000, month=1),
                               dict(year=5000, month=3))
        
        # jun 5000 to jan -> jun 1 5000 to jan 31 5001
        self.assert_unamb_span(p.date_from("jun 5000 to jan", basedate),
                               dict(year=5000, month=6),
                               dict(year=5001, month=1))
        
        # oct 2010 to feb -> oct 1 2010 to feb 28 2011
        self.assert_unamb_span(p.date_from("oct 2010 to feb"),
                               dict(year=2010, month=10),
                               dict(year=2011, month=2))
        
        self.assert_unamb_span(p.date_from("5pm to 3am", basedate),
                               dict(year=2010, month=9, day=20, hour=17),
                               dict(year=2010, month=9, day=21, hour=3))
        
        self.assert_unamb_span(p.date_from("5am to 3 am tomorrow", basedate),
                               dict(year=2010, month=9, day=20, hour=5),
                               dict(year=2010, month=9, day=21, hour=3))
        
        self.assert_unamb_span(p.date_from("3am to 5 pm tomorrow", basedate),
                               dict(year=2010, month=9, day=21, hour=3),
                               dict(year=2010, month=9, day=21, hour=17))
        
        self.assert_unamb_span(p.date_from("-2hrs to +20min", basedate),
                               dict(year=2010, month=9, day=20, hour=13, minute=16, second=6, microsecond=454000),
                               dict(year=2010, month=9, day=20, hour=15, minute=36, second=6, microsecond=454000))
        
        # Swap
        self.assert_unamb_span(p.date_from("oct 25 2009 to feb 14 2008", basedate),
                               dict(year=2008, month=2, day=14),
                               dict(year=2009, month=10, day=25))
    
        self.assert_unamb_span(p.date_from("oct 25 5000 to tomorrow", basedate),
                               dict(year=2010, month=9, day=21),
                               dict(year=5000, month=10, day=25))
        
        
        


if __name__ == '__main__':
    unittest.main()






from __future__ import with_statement
import unittest

import threading, time

from whoosh.support.filelock import try_for
from whoosh.util import length_to_byte, byte_to_length
from whoosh.support.testing import TempStorage


class TestMisc(unittest.TestCase):
    def test_filelock_simple(self):
        with TempStorage("simplefilelock") as st:
            lock1 = st.lock("testlock")
            lock2 = st.lock("testlock")
            self.assertFalse(lock1 is lock2)
            
            self.assertTrue(lock1.acquire())
            self.assertTrue(st.file_exists("testlock"))
            self.assertFalse(lock2.acquire())
            lock1.release()
            self.assertTrue(lock2.acquire())
            self.assertFalse(lock1.acquire())
            lock2.release()
        
    def test_threaded_filelock(self):
        with TempStorage("threadedfilelock") as st:
            lock1 = st.lock("testlock")
            result = []
            
            # The thread function tries to acquire the lock and then quits
            def fn():
                lock2 = st.lock("testlock")
                gotit = try_for(lock2.acquire, 1.0, 0.1)
                if gotit:
                    result.append(True)
                    lock2.release()
            t = threading.Thread(target=fn)
            
            # Acquire the lock in this thread
            lock1.acquire()
            # Start the other thread trying to acquire the lock
            t.start()
            # Wait for a bit
            time.sleep(0.15)
            # Release the lock
            lock1.release()
            # Wait for the other thread to finish
            t.join()
            # If the other thread got the lock, it should have appended True to the
            # "results" list.
            self.assertEqual(result, [True])
        
    def test_length_byte(self):
        source = range(11)
        xform = [length_to_byte(n) for n in source]
        result = [byte_to_length(n) for n in xform]
        self.assertEqual(source, result)
        
    def test_lru_cache(self):
        from whoosh.util import lru_cache
        
        @lru_cache(5)
        def test(n):
            return n * 2
        
        result = [test(n) for n in (1, 2, 3, 4, 5, 4, 3, 2, 10, 1)]
        self.assertEqual(result, [2, 4, 6, 8, 10, 8, 6, 4, 20, 2])
        self.assertEqual(test.cache_info(), (3, 7, 5, 5))
        test.cache_clear()
        self.assertEqual(test.cache_info(), (0, 0, 5, 0))
        


if __name__ == '__main__':
    unittest.main()
